from typing import List, Optional, Union
from ddd_objects.domain.exception import ParameterError
from pydantic import BaseModel

class BanInstanceTypeRequestDTO(BaseModel):
    instance_type: str
    region_id: str
    zone_id : str
    duration: Optional[int] = None

class BannedInstanceTypeResponseDTO(BaseModel):
    instance_types: Optional[List[str]]
    region_id: str
    zone_id: Optional[str]

class ConditionDTO(BaseModel):
    min_cpu_num: Optional[int] = 1
    max_cpu_num: Optional[int] = 1
    min_memory_size: Optional[int] = 1
    max_memory_size: Optional[int] = 1
    min_gpu_num: Optional[int] = None
    max_gpu_num: Optional[int] = None
    min_gpu_memory_size: Optional[int] = None
    max_gpu_memory_size: Optional[int] = None

class InstanceUserSettingDTO(BaseModel):
    name: str
    password: str = 'Abcd1234'
    amount: int = 1
    image_id: Optional[str]
    region_id: str
    internet_pay_type: Optional[str]
    bandwidth_in: int = 200
    bandwidth_out: int = 1
    user_data: Optional[str] = None
    disk_size: int = 20
    key_name: str
    exclude_instance_types: List[str]=[]
    inner_connection: bool = True

class InstanceCreationRequestDTO(BaseModel):
    instance_user_setting: InstanceUserSettingDTO
    condition: ConditionDTO
    priority: int = 3
    timeout: int = 400

class InstanceInfoDTO(BaseModel):
    id: str
    instance_type: str
    create_time: str
    name: str
    hostname: str
    pay_type: str
    public_ip: List[str]
    private_ip: Optional[str]
    os_name: str
    price: float
    image_id: str
    region_id: str
    zone_id: str
    internet_pay_type: str
    bandwidth_in: int
    bandwidth_out: int
    security_group_id: List[str]
    instance_expired_time: Optional[str]
    auto_release_time: Optional[str]
    status: str
    key_name: str
    _life_time: int = 5

class InstanceCreationItemDTO(BaseModel):
    id: str
    instance_creation_request: Optional[InstanceCreationRequestDTO]
    status: str
    creation_time: str
    details: Optional[List[InstanceInfoDTO]] = None
    entry_time: Optional[str] = None
    exit_time: Optional[str] = None
    exception: Optional[str] = None
    _life_time: int=86400

class ReleaseInstanceInfoDTO(BaseModel):
    region_id: str
    instance_id: str

class CommandContextDTO(BaseModel):
    connection: str='smart'
    module_path: str='/usr/local/bin/ansible'
    forks: int= 1
    timeout: int=300 
    remote_user: str='root'
    ask_pass: bool = False 
    ssh_extra_args: Optional[str] = None
    sftp_extra_args: Optional[str] = None
    scp_extra_args: Optional[str] = None 
    ask_value_pass: bool = False
    listhosts: bool = False 
    listtasks: bool = False 
    listtags: bool = False
    syntax: bool = False
    ssh_common_args='-o StrictHostKeyChecking=no'
    become: Optional[str] = None
    become_method: Optional[str] = None
    become_user: str = 'root' 
    check: bool = False 
    diff: bool = False
    verbosity: int = 1

class CommandSettingDetailDTO(BaseModel):
    command: Union[str, dict]
    module: str = 'shell'
    retries: int = 2
    timeout: int = 3
    delay: float = 0.1

class CommandHostDTO(BaseModel):
    ip: str
    hostname: Optional[str] = None
    port: int = 22
    username: str = 'root'

class CommandRequestDTO(BaseModel):
    commands: Union[List[str], List[CommandSettingDetailDTO]]
    hosts: Union[List[str], List[CommandHostDTO]]
    context: Optional[CommandContextDTO]=None
    priority: int = 3
    timeout: int = 3

    def __init__(self, **data) -> None:
        commands = data['commands']
        hosts = data['hosts']
        if len(commands)==0:
            raise ParameterError('No command is set when creating CommandRequestDTO object.')
        if len(hosts)==0:
            raise ParameterError('No host is set when creating CommandRequestDTO object.')
        if isinstance(commands[0], str):
            data['commands'] = [CommandSettingDetailDTO(command=c) for c in commands]
        if isinstance(hosts[0], str):
            data['hosts'] = [CommandHostDTO(ip=h) for h in hosts]
        super().__init__(**data)
    
class CommandResponseDTO(BaseModel):
    status: str
    ip: str
    message: Optional[str] = None
    exception: Optional[str] = None
    stdout: Optional[str] = None
    cmd: Optional[str] = None
    start_time: Optional[str] = None
    end_time: Optional[str] = None
    delta_time: Optional[str] = None
    stderr: Optional[str] = None
    entry_time: Optional[str] = None
    exit_time: Optional[str] = None

class CommandItemDTO(BaseModel):
    id: str
    command_request: Optional[CommandRequestDTO]
    creation_time: str
    status: str
    details: Optional[List[CommandResponseDTO]] = None
    exception: Optional[str] = None
    _life_time: int = 600

class DNSRecordDTO(BaseModel):
    domain_name: str
    subdomain: str
    value: str
    id: Optional[str]=None
    weight: Optional[int]=None
    dns_type: str='A'
    ttl: int=600
    priority: Optional[int]=None
    line: Optional[str]=None



from ddd_objects.domain.entity import Entity
from typing import Optional, Any, Callable
from .value_obj_ext import PipelineContext, PipelineIndicator
from .value_obj import Name

class _PipelineStage(Entity):
    def __init__(
        self,
        stage_name: str,
        action: Callable,
        allow_failure: bool=False
    ) -> None:
        self.stage_name = stage_name
        self.action = action
        self.allow_failure = allow_failure

class PipelineActionOutput(Entity):
    def __init__(
        self,
        context: PipelineContext,
        indicator: PipelineIndicator=PipelineIndicator(),
        jump_stage: Optional[Name] = None,
        output: Optional[Any]=None
    ) -> None:
        self.context = context
        self.indicator = indicator
        self.jump_stage = jump_stage
        self.output = output



import datetime, json
from dataclasses import dataclass
from re import I
from typing import Dict, List, Optional
from ddd_objects.domain.entity import Entity, ExpiredEntity
from ddd_objects.domain.exception import ValueError
from .value_obj import (
    CommandException,
    InstanceException,
    InstanceName,
    KeyName,
    U,
    CommandID,
    CommandModule,
    CommandPriority,
    CommandResponseStatus,
    CommandStatus,
    CreationTime,
    DeltaTime,
    InstanceCreationID,
    InstanceCreationPriority,
    InstanceCreationStatus,
    Price,
    InstanceID,
    Command,
    Bandwidth,
    Password,
    SecurityGroupID,
    Bool,
    DomainName,
    MinCPUNumber,
    MaxCPUNumber,
    Size,
    Path,
    InstanceTypeID,
    InstanceStatus,
    Number,
    Subdomain,
    Value,
    ID,
    Line,
    Data,
    Weight,
    MaxGPUNumber,
    Text,
    ImageID,
    DateTime,
    ZoneID,
    DNSType,
    MinMemorySize,
    Priority,
    Name,
    IP,
    MinGPUMemorySize,
    MaxGPUMemorySize,
    Type,
    MinGPUNumber,
    MaxMemorySize,
    RegionID
)

class BanInstanceTypeRequest(Entity):
    def __init__(
        self, 
        instance_type: InstanceTypeID, 
        region_id: RegionID,
        zone_id: ZoneID,
        duration: Optional[Number]=None
    ) -> None:
        self.instance_type = instance_type
        self.region_id = region_id
        self.zone_id = zone_id
        self.duration = duration
        super().__init__()

class BannedInstanceTypeResponse(Entity):
    def __init__(
        self, 
        instance_types: List[InstanceTypeID],
        region_id: RegionID,
        zone_id: Optional[ZoneID]

    ):
        self.instance_types = instance_types
        self.region_id = region_id
        self.zone_id = zone_id

class BannedInstanceTypeItem(ExpiredEntity):
    def __init__(
        self, 
        ban_instance_type_request:BanInstanceTypeRequest,
        _life_time: Number = Number(86400)
    ) -> None:
        self.ban_instance_type_request = ban_instance_type_request
        duration = ban_instance_type_request.duration
        if duration is None:
            self.ban_instance_type_request.duration = _life_time
        else:
            _life_time = duration
        super().__init__(_life_time=_life_time)

class Condition(Entity):
    def __init__(
        self,
        min_cpu_num: MinCPUNumber=MinCPUNumber(1),
        max_cpu_num: MaxCPUNumber=MaxCPUNumber(1),
        min_memory_size: MinMemorySize=MinMemorySize(1),
        max_memory_size: MaxMemorySize=MaxMemorySize(1),
        min_gpu_num: Optional[MinGPUNumber]=None,
        max_gpu_num: Optional[MaxGPUNumber]=None,
        min_gpu_memory_size: Optional[MinGPUMemorySize]=None,
        max_gpu_memory_size: Optional[MaxGPUMemorySize]=None
    ):
        
        if max_cpu_num.get_value()< min_cpu_num.get_value():
            raise ValueError('max_cpu_num should greater than min_cpu_num')
        if max_memory_size.get_value()<min_memory_size.get_value():
            raise ValueError('max_memory_size should greater than min_memory_size')
        if max_gpu_num is not None and min_gpu_num is not None and \
            max_gpu_num.get_value()< min_gpu_num.get_value():
            raise ValueError('max_gpu_num should greater than min_gpu_num')
        if max_gpu_memory_size is not None and min_gpu_memory_size is not None and \
            max_gpu_memory_size.get_value()<min_gpu_memory_size.get_value():
            raise ValueError('max_gpu_memory_size should greater than min_gpu_memory_size')
        
        self.min_cpu_num = min_cpu_num
        self.max_cpu_num = max_cpu_num
        self.min_memory_size = min_memory_size
        self.max_memory_size = max_memory_size
        self.min_gpu_num = min_gpu_num
        self.max_gpu_num = max_gpu_num
        self.min_gpu_memory_size = min_gpu_memory_size
        self.max_gpu_memory_size = max_gpu_memory_size

    def __str__(self) -> str:
        return json.dumps({
            'min_cpu_num': self.min_cpu_num.get_value(),
            'max_cpu_num': self.max_cpu_num.get_value(),
            'min_memory_size': self.min_memory_size.get_value(),
            'max_memory_size': self.max_memory_size.get_value(),
            'min_gpu_num': None if self.min_gpu_num is None else self.min_gpu_num.get_value(),
            'max_gpu_num': None if self.max_gpu_num is None else self.max_gpu_num.get_value(),
            'min_gpu_memory_size': None if self.min_gpu_memory_size is None else self.min_gpu_memory_size.get_value(),
            'max_gpu_memory_size': None if self.max_gpu_memory_size is None else self.max_gpu_memory_size.get_value()
        }).replace('\\', '')

@dataclass
class InstanceInfo(Entity):
    id: InstanceID
    instance_type: InstanceTypeID 
    create_time: DateTime 
    name: Name
    hostname: Name
    pay_type: Type
    public_ip: List[IP]
    private_ip: Optional[IP]
    os_name: Name
    price: Price
    image_id: ImageID
    region_id: RegionID
    zone_id: ZoneID
    internet_pay_type: Type
    bandwidth_in: Bandwidth
    bandwidth_out: Bandwidth
    status: InstanceStatus
    key_name: KeyName
    security_group_id: List[SecurityGroupID]
    instance_expired_time: Optional[DateTime] = None
    auto_release_time: Optional[DateTime] = None
    _life_time: Number = Number(5)


@dataclass
class InstanceSetting(Entity):
    name: Name
    region_id: RegionID
    zone_id: ZoneID
    security_group_id: SecurityGroupID
    instance_type: InstanceTypeID
    password: Password
    amount: Number
    spot_price: Price
    image_id: ImageID
    internet_pay_type: Type
    bandwidth_in: Bandwidth
    bandwidth_out: Bandwidth
    user_data: Data
    disk_size: Size
    key_name: KeyName

class InstanceUserSetting(Entity):
    def __init__(
        self,
        name: InstanceName = InstanceName('test'),
        password: Password=Password('1234Abcd'),
        amount: Number = Number(1),
        image_id: Optional[ImageID] = ImageID('centos_7_9_x64_20G_alibase_20220727.vhd'),
        region_id: RegionID = RegionID('cn-zhangjiakou'),
        internet_pay_type: Optional[Type] = None,
        bandwidth_in: Bandwidth = Bandwidth(200),
        bandwidth_out: Bandwidth = Bandwidth(1),
        user_data: Optional[Data] = None,
        disk_size: Size = Size(20),
        key_name: KeyName = KeyName('ansible'),
        exclude_instance_types: List[InstanceTypeID] = [],
        inner_connection: Bool = Bool(True)
    ) -> None:
        self.name = name
        self.password = password
        self.amount = amount
        self.image_id = image_id
        self.region_id = region_id
        self.internet_pay_type = internet_pay_type
        self.bandwidth_in = bandwidth_in
        self.bandwidth_out = bandwidth_out
        self.user_data = user_data
        self.disk_size = disk_size
        self.key_name = key_name
        self.exclude_instance_types = exclude_instance_types
        self.inner_connection = inner_connection
    
    def __str__(self) -> str:
        return json.dumps({
            'name': self.name.get_value(),
            'password': self.password.get_value(),
            'amount': self.amount.get_value(),
            'image_id': None if self.image_id is None else self.image_id.get_value(),
            'region_id': self.region_id.get_value(),
            'internet_pay_type': None if self.internet_pay_type is None else self.internet_pay_type.get_value(),
            'bandwidth_in': self.bandwidth_in.get_value(),
            'bandwidth_out': self.bandwidth_out.get_value(),
            'user_data': None if self.user_data is None else self.user_data.get_value(),
            'disk_size': self.disk_size.get_value(),
            'key_name': self.key_name.get_value(),
            'exclude_instance_types': [t.get_value() for t in self.exclude_instance_types],
            'inner_connectioin': self.inner_connection.get_value()
        })

class InstanceCreationRequest(Entity):
    def __init__(
        self, 
        instance_user_setting: InstanceUserSetting, 
        condition: Condition,
        priority: InstanceCreationPriority,
        timeout: Number
    ) -> None:
        self.instance_user_setting = instance_user_setting
        self.condition = condition
        self.priority = priority
        self.timeout = timeout
    
    def __str__(self) -> str:
        return json.dumps({
            'instance_user_setting': str(self.instance_user_setting),
            'condition': str(self.condition)
        }).replace('\\', '')

class InstanceCreationItem(ExpiredEntity):
    def __init__(
        self, 
        id: InstanceCreationID,
        instance_creation_request: InstanceCreationRequest,
        creation_time:CreationTime = CreationTime(),
        status: InstanceCreationStatus = InstanceCreationStatus(),
        details: Optional[List[InstanceInfo]] = None,
        entry_time: Optional[DateTime] = None,
        exit_time: Optional[DateTime] = None,
        _life_time: Number = Number(86400)
    ) -> None:
        self.id = id
        self.instance_creation_request = instance_creation_request
        self.status = status
        self.creation_time=creation_time
        self.details = details
        self.entry_time = entry_time
        self.exit_time = exit_time
        self._life_time = _life_time
        super().__init__(_life_time=_life_time)

    def is_sending_timeout(self, timeout=30):
        if self.status.is_sent():
            diff = datetime.datetime.utcnow()-self.status.changed_time
            return diff.seconds>timeout
        return False

    def __str__(self) -> str:
        return json.dumps({
            'id': self.id.get_value(),
            'instance_creation_request': str(self.instance_creation_request),
            'creation_time': str(self.creation_time),
            'status': self.status.get_value(),
            'details': None if self.details is None else [str(d) for d in self.details],
            '_life_time': self._life_time.get_value()
        })
    

class CommandContext(Entity):
    def __init__(
        self,
        connection: U,
        module_path: Path,
        forks: Number,
        timeout: Number,
        remote_user: Name,
        ask_pass: Bool,
        ssh_extra_args: Optional[U],
        sftp_extra_args: Optional[U],
        scp_extra_args: Optional[U],
        ask_value_pass: Bool,
        listhosts: Bool,
        listtasks: Bool,
        listtags: Bool,
        syntax: Bool,
        ssh_common_args: U,
        become: Optional[U],
        become_method: Optional[str],
        become_user: Name,
        check: Bool,
        diff: Bool,
        verbosity: Number

    ):
        self.connection = connection
        self.module_path = module_path
        self.forks = forks
        self.timeout = timeout
        self.remote_user = remote_user
        self.ask_pass = ask_pass
        self.ssh_extra_args = ssh_extra_args
        self.sftp_extra_args = sftp_extra_args
        self.scp_extra_args = scp_extra_args
        self.ask_value_pass = ask_value_pass
        self.listhosts = listhosts
        self.listtasks = listtasks
        self.listtags = listtags
        self.syntax = syntax
        self.ssh_common_args = ssh_common_args
        self.become = become
        self.become_method = become_method
        self.become_user = become_user
        self.check = check
        self.diff = diff
        self.verbosity = verbosity

    def __str__(self) -> str:
        return json.dumps({
            'connection': self.connection.get_value(),
            'module_path': self.module_path.get_value(),
            'forks': self.forks.get_value(),
            'timeout': self.timeout.get_value(),
            'remote_user': self.remote_user.get_value(),
            'ask_pass': self.ask_pass.get_value(),
            'ssh_extra_args': None if self.ssh_extra_args is None else self.ssh_extra_args.get_value(),
            'sftp_extra_args': None if self.sftp_extra_args is None else self.sftp_extra_args.get_value(),
            'scp_extra_args': None if self.scp_extra_args is None else self.scp_extra_args.get_value(),
            'ask_value_pass': self.ask_value_pass.get_value(),
            'listhosts': self.listhosts.get_value(),
            'listtasks': self.listtasks.get_value(),
            'listtags': self.listtags.get_value(),
            'syntax': self.syntax.get_value(),
            'ssh_common_args': self.ssh_common_args.get_value(),
            'become': None if self.become is None else self.become.get_value(),
            'become_method': None if self.become_method is None else self.become_method.get_value(),
            'become_user': self.become_user.get_value(),
            'check': self.check.get_value(),
            'diff': self.diff.get_value(),
            'verbosity': self.verbosity.get_value()
        }).replace('\\', '')

class CommandSettingDetail(Entity):
    def __init__(
        self,
        command: Command,
        module: CommandModule=CommandModule('shell'),
        retries: Number=Number(2),
        timeout: Number=Number(3),
        delay: Number=Number(0.1)
    ) -> None:
        self.command = command
        self.module = module
        self.retries = retries
        self.timeout = timeout
        self.delay = delay
    
    def __str__(self) -> str:
        return json.dumps({
            'command': self.command.get_value(),
            'module': self.module.get_value(),
            'retries': self.retries.get_value(),
            'timeout': self.timeout.get_value(),
            'delay': self.delay.get_value()
        }).replace('\\', '')

class CommandHost(Entity):
    def __init__(
        self,
        ip: IP,
        hostname: Optional[Name],
        port: Number,
        username: Name
    ) -> None:
        self.ip = ip
        self.hostname = hostname
        self.port = port
        self.username = username

    def __str__(self) -> str:
        return json.dumps({
            'ip': self.ip.get_value(),
            'hostname': None if self.hostname is None else self.hostname.get_value(),
            'port': self.port.get_value(),
            'username': self.username.get_value()
        }).replace('\\', '')

class CommandRequest(Entity):
    def __init__(
        self,
        commands: List[CommandSettingDetail],
        hosts: List[CommandHost],
        context: Optional[CommandContext],
        priority: CommandPriority,
        timeout: Number,
    ) -> None:
        self.commands = commands
        self.hosts = hosts
        self.context = context
        self.priority = priority
        self.timeout = timeout

    def __str__(self) -> str:
        return json.dumps({
            'commands': [str(c) for c in self.commands],
            'hosts': [str(h) for h in self.hosts],
            'context': None if self.context is None else str(self.context),
            'priority': self.priority.get_value(),
            'timeout': self.timeout.get_value()
        }).replace('\\', '')

class CommandResponse(Entity):
    def __init__(
        self,
        status: CommandResponseStatus,
        ip: IP,
        message: Text,
        exception: Optional[Text],
        stdout: Optional[Text],
        cmd: Optional[Text],
        start_time: Optional[DateTime],
        end_time: Optional[DateTime],
        delta_time: Optional[DeltaTime],
        stderr: Optional[Text],
        entry_time: Optional[DateTime],
        exit_time: Optional[DateTime]
    ) -> None:
        self.status = status
        self.ip = ip
        self.message = message
        self.exception = exception
        self.stdout = stdout
        self.cmd = cmd
        self.start_time = start_time
        self.end_time = end_time
        self.delta_time = delta_time
        self.stderr = stderr
        self.entry_time = entry_time
        self.exit_time = exit_time

    def __str__(self) -> str:
        return json.dumps({
            'status': self.status.get_value(),
            'ip': self.ip.get_value(),
            'message': self.message.get_value(),
            'exception': None if self.exception is None else self.exception.get_value(),
            'stdout': None if self.stdout is None else self.stdout.get_value(),
            'cmd': None if self.cmd is None else self.cmd.get_value(),
            'start_time': None if self.start_time is None else self.start_time.get_value(),
            'end_time': None if self.end_time is None else self.end_time.get_value(),
            'delta_time': None if self.delta_time is None else self.delta_time.get_value(),
            'stderr': None if self.stderr is None else self.stderr.get_value(),
            'entry_time': None if self.entry_time is None else self.entry_time.get_value(),
            'exit_time': None if self.exit_time is None else self.exit_time.get_value()
        }).replace('\\', '')

class CommandItem(ExpiredEntity):
    def __init__(
        self, 
        id: CommandID,
        command_request: Optional[CommandRequest],
        creation_time:CreationTime = CreationTime(datetime.datetime.utcnow()),
        status: CommandStatus = CommandStatus(),
        details: Optional[List[CommandResponse]] = None,
        exception: Optional[CommandException] = None,
        _life_time: Number = Number(600)
    ) -> None:
        self.id = id
        self.command_request = command_request
        self.status = status
        self.creation_time=creation_time
        self.details = details
        self._life_time = _life_time
        self.exception = exception
        super().__init__(_life_time=_life_time)

    def is_sending_timeout(self, timeout=30):
        if self.status.is_sent():
            diff = datetime.datetime.utcnow()-self.status.changed_time
            return diff.seconds>timeout
        return False

    def __str__(self) -> str:
        return json.dumps({
            'id': self.id.get_value(),
            'command_request': None if self.command_request is None else str(self.command_request),
            'creation_time': str(self.creation_time),
            'status': self.status.get_value(),
            'details': None if self.details is None else [str(d) for d in self.details],
            'exception': None if self.exception is None else self.exception.get_value(),
        }).replace('\\', '')


class InstanceCreationItem(ExpiredEntity):
    def __init__(
        self, 
        id: InstanceCreationID,
        instance_creation_request: InstanceCreationRequest,
        creation_time:CreationTime = CreationTime(datetime.datetime.utcnow()),
        status: InstanceCreationStatus = InstanceCreationStatus(),
        details: Optional[List[InstanceInfo]] = None,
        entry_time: Optional[DateTime] = None,
        exit_time: Optional[DateTime] = None,
        exception: Optional[InstanceException] = None,
        _life_time: Number = Number(86400)
    ) -> None:
        self.id = id
        self.instance_creation_request = instance_creation_request
        self.status = status
        self.creation_time=creation_time
        self.details = details
        self._life_time = _life_time
        self.entry_time = entry_time
        self.exit_time = exit_time
        self.exception = exception
        super().__init__(_life_time=_life_time)

    def is_sending_timeout(self, timeout=30):
        if self.status.is_sent():
            diff = datetime.datetime.utcnow()-self.status.changed_time
            return diff.seconds>timeout
        return False

    def __str__(self) -> str:
        return json.dumps({
            'id': self.id.get_value(),
            'instance_creation_request': None if self.instance_creation_request is None else str(self.instance_creation_request),
            'status': self.status.get_value(),
            'creation_time': str(self.creation_time),
            'details': None if self.details is None else [str(d) for d in self.details],
            '_life_time': self._life_time.get_value(),
            'entry_time': None if self.entry_time is None else self.entry_time.get_value(),
            'exit_time': None if self.exit_time is None else self.exit_time.get_value()
        }).replace('\\', '')


class DNSRecord(Entity):
    def __init__(
        self,
        domain_name: DomainName,
        subdomain: Subdomain,
        value: Value,
        id: Optional[ID] = None,
        weight: Optional[Weight] = None,
        dns_type: DNSType = DNSType('A'),
        ttl: Number = Number(600),
        priority: Optional[Priority] = None,
        line: Optional[Line] = None
    ):
        self.domain_name=domain_name
        self.subdomain=subdomain
        self.value=value
        self.id=id
        self.weight=weight
        self.dns_type=dns_type
        self.ttl=ttl
        self.priority=priority
        self.line=line




from typing import List
from ddd_objects.application.assembler import Assembler
from .dto import (
    BanInstanceTypeRequestDTO,
    BannedInstanceTypeResponseDTO,
    CommandContextDTO,
    CommandHostDTO,
    CommandItemDTO,
    CommandRequestDTO,
    CommandResponseDTO,
    CommandSettingDetailDTO,
    InstanceCreationItemDTO,
    InstanceCreationRequestDTO,
    ConditionDTO,
    InstanceUserSettingDTO,
    DNSRecordDTO,
    InstanceInfoDTO
)
from ..domain.entity import (
    BanInstanceTypeRequest,
    BannedInstanceTypeResponse,
    CommandContext,
    CommandHost,
    CommandItem,
    CommandRequest,
    CommandResponse,
    CommandSettingDetail,
    InstanceCreationItem,
    InstanceCreationRequest,
    InstanceInfo,
    Condition,
    DNSRecord,
    InstanceUserSetting,
)
from ..domain.value_obj import (
    CommandException,
    InstanceException,
    InstanceName,
    KeyName,
    U,
    Command,
    CommandModule,
    CommandPriority,
    CommandResponseStatus,
    CommandStatus,
    CreationTime,
    DeltaTime,
    InstanceCreationID,
    InstanceCreationPriority,
    InstanceCreationStatus,
    Price,
    InstanceID,
    Bandwidth,
    Password,
    Bool,
    SecurityGroupID,
    DomainName,
    MinCPUNumber,
    Text,
    MaxCPUNumber,
    Size,
    Path,
    InstanceTypeID,
    Number,
    Status,
    Subdomain,
    Value,
    ID,
    Line,
    Data,
    Weight,
    ImageID,
    DateTime,
    ZoneID,
    DNSType,
    MinMemorySize,
    Priority,
    Name,
    IP,
    MinGPUMemorySize,
    MaxGPUMemorySize,
    Type,
    MinGPUNumber,
    MaxGPUNumber,
    MaxMemorySize,
    RegionID
)
from .dto import (
    DNSRecordDTO,
    ConditionDTO,
    InstanceInfoDTO,
    InstanceUserSettingDTO
)

class ConditionAssembler(Assembler):
    @Assembler.assemble_none
    def to_entity(self, dto: ConditionDTO):
        return Condition(
            min_cpu_num=MinCPUNumber(dto.min_cpu_num),
            max_cpu_num=MaxCPUNumber(dto.max_cpu_num),
            min_memory_size=MinMemorySize(dto.min_memory_size),
            max_memory_size=MaxMemorySize(dto.max_memory_size),
            min_gpu_num=None if dto.min_gpu_num is None else MinGPUNumber(dto.min_gpu_num),
            max_gpu_num=None if dto.max_gpu_num is None else MaxGPUNumber(dto.max_gpu_num),
            min_gpu_memory_size=None if dto.min_gpu_memory_size is None else MinGPUMemorySize(dto.min_gpu_memory_size),
            max_gpu_memory_size=None if dto.max_gpu_memory_size is None else MaxGPUMemorySize(dto.max_gpu_memory_size)
        )
    def to_dto(self, x: Condition):
        return ConditionDTO(
            min_cpu_num=x.min_cpu_num.get_value(),
            max_cpu_num=x.max_cpu_num.get_value(),
            min_memory_size=x.min_memory_size.get_value(),
            max_memory_size=x.max_memory_size.get_value(),
            min_gpu_num=None if x.min_gpu_num is None else x.min_gpu_num.get_value(),
            max_gpu_num=None if x.max_gpu_num is None else x.max_gpu_num.get_value(),
            min_gpu_memory_size=None if x.min_gpu_memory_size is None else x.min_gpu_memory_size.get_value(),
            max_gpu_memory_size=None if x.max_gpu_memory_size is None else x.max_gpu_memory_size.get_value()
        )
condition_assembler = ConditionAssembler()

class InstanceUserSettingAssembler(Assembler):
    @Assembler.assemble_none
    def to_entity(self, dto: InstanceUserSettingDTO):
        return InstanceUserSetting(
            name = InstanceName(dto.name),
            password=None if dto.password is None else Password(dto.password),
            amount=Number(dto.amount),
            image_id = None if dto.image_id is None else ImageID(dto.image_id),
            region_id=RegionID(dto.region_id),
            internet_pay_type=None if dto.internet_pay_type is None else Type(dto.internet_pay_type),
            bandwidth_in=Bandwidth(dto.bandwidth_in),
            bandwidth_out=Bandwidth(dto.bandwidth_out),
            user_data=None if dto.user_data is None else Data(dto.user_data),
            disk_size=Size(dto.disk_size),
            key_name = KeyName(dto.key_name),
            exclude_instance_types=[InstanceTypeID(x) for x in dto.exclude_instance_types],
            inner_connection = Bool(dto.inner_connection)
        )
    def to_dto(self, x: InstanceUserSetting):
        return InstanceUserSettingDTO(
            name = x.name.get_value(),
            password = None if x.password is None else x.password.get_value(),
            amount=x.amount.get_value(),
            image_id = None if x.image_id is None else x.image_id.get_value(),
            region_id= x.region_id.get_value(),
            internet_pay_type=None if x.internet_pay_type is None else x.internet_pay_type.get_value(),
            bandwidth_in= x.bandwidth_in.get_value(),
            bandwidth_out= x.bandwidth_out.get_value(),
            user_data=None if x.user_data is None else x.user_data.get_value(),
            disk_size=x.disk_size.get_value(),
            key_name = x.key_name.get_value(),
            exclude_instance_types=[m.get_value() for m in x.exclude_instance_types],
            inner_connection = x.inner_connection.get_value()
        )
instance_user_setting_assembler = InstanceUserSettingAssembler()

class InstanceInfoAssembler(Assembler):
    @Assembler.assemble_none
    def to_entity(self, dto: InstanceInfoDTO):
        return InstanceInfo(
            id = InstanceID(dto.id),
            instance_type=InstanceTypeID(dto.instance_type),
            create_time=DateTime(dto.create_time),
            name=Name(dto.name),
            hostname = Name(dto.hostname),
            pay_type = Type(dto.pay_type),
            public_ip = [IP(ip) for ip in dto.public_ip],
            private_ip = None if dto.private_ip is None else IP(dto.private_ip),
            os_name = Name(dto.os_name),
            price = Price(dto.price),
            image_id=ImageID(dto.image_id),
            region_id=RegionID(dto.region_id),
            zone_id = ZoneID(dto.zone_id),
            internet_pay_type = Type(dto.internet_pay_type),
            bandwidth_in = Bandwidth(dto.bandwidth_in),
            bandwidth_out = Bandwidth(dto.bandwidth_out),
            security_group_id = [SecurityGroupID(m) for m in dto.security_group_id],
            instance_expired_time = DateTime(dto.instance_expired_time),
            auto_release_time = DateTime(dto.auto_release_time),
            status = Status(dto.status),
            key_name=KeyName(dto.key_name)
        )
    @Assembler.assemble_none
    def to_dto(self, info: InstanceInfo):
        return InstanceInfoDTO(
            id=info.id.get_value(),
            instance_type=info.instance_type.get_value(),
            create_time=info.create_time.get_value(),
            name=info.name.get_value(),
            hostname=info.hostname.get_value(),
            pay_type=info.pay_type.get_value(),
            public_ip=[ip.get_value() for ip in info.public_ip],
            private_ip=None if info.private_ip is None else info.private_ip.get_value(),
            os_name=info.os_name.get_value(),
            price=info.price.get_value(),
            image_id=info.image_id.get_value(),
            region_id=info.region_id.get_value(),
            zone_id=info.zone_id.get_value(),
            internet_pay_type=info.internet_pay_type.get_value(),
            bandwidth_in=info.bandwidth_in.get_value(),
            bandwidth_out=info.bandwidth_out.get_value(),
            security_group_id=[id.get_value() for id in info.security_group_id],
            instance_expired_time=info.instance_expired_time.get_value(),
            auto_release_time=info.auto_release_time.get_value(),
            status=info.status.get_value(),
            key_name=info.key_name.get_value()
        )
instance_info_assembler = InstanceInfoAssembler()

class InstanceCreationRequestAssembler(Assembler):
    def to_entity(self, dto: InstanceCreationRequestDTO):
        return InstanceCreationRequest(
            instance_user_setting=instance_user_setting_assembler.to_entity(dto.instance_user_setting),
            condition=condition_assembler.to_entity(dto.condition),
            priority=InstanceCreationPriority(dto.priority),
            timeout=Number(dto.timeout) 
        )
    def to_dto(self, x: InstanceCreationRequest):
        return InstanceCreationRequestDTO(
            instance_user_setting = instance_user_setting_assembler.to_dto(x.instance_user_setting),
            condition = condition_assembler.to_dto(x.condition),
            priority=x.priority.get_value(),
            timeout=x.timeout.get_value()
        )
instance_creation_request_assembler = InstanceCreationRequestAssembler()

class InstanceCreationItemAssembler(Assembler):
    def to_entity(self, dto: InstanceCreationItemDTO):
        return InstanceCreationItem(
            id = InstanceCreationID(dto.id),
            instance_creation_request=None if dto.instance_creation_request is None else instance_creation_request_assembler.to_entity(dto.instance_creation_request),
            creation_time=CreationTime(dto.creation_time),
            status = InstanceCreationStatus(dto.status),
            details = None if dto.details is None else [instance_info_assembler.to_entity(m) for m in dto.details],
            entry_time = None if dto.entry_time is None else DateTime(dto.entry_time),
            exit_time = None if dto.exit_time is None else DateTime(dto.exit_time),
            exception = None if dto.exception is None else InstanceException(dto.exception),
            _life_time = Number(dto._life_time)
        )
    
    def to_dto(self, x: InstanceCreationItem):
        return InstanceCreationItemDTO(
            id = x.id.get_value(),
            instance_creation_request = None if x.instance_creation_request is None else instance_creation_request_assembler.to_dto(x.instance_creation_request),
            creation_time = x.creation_time.get_value().replace(microsecond=0).isoformat(),
            status = x.status.get_value(),
            details = None if x.details is None else [instance_info_assembler.to_dto(m) for m in x.details],
            entry_time = None if x.entry_time is None else x.entry_time.get_value(),
            exit_time = None if x.exit_time is None else x.exit_time.get_value(),
            exception = None if x.exception is None else x.exception.get_value(),
            _life_time = x._life_time.get_value()
        )
instance_creation_item_assembler = InstanceCreationItemAssembler()

class BanInstanceTypeRequestAssembler(Assembler):
    def to_entity(self, dto: BanInstanceTypeRequestDTO):
        return BanInstanceTypeRequest(
            instance_type = InstanceTypeID(dto.instance_type),
            region_id= RegionID(dto.region_id),
            zone_id=ZoneID(dto.zone_id),
            duration = None if dto.duration is None else Number(dto.duration)
        )

    def to_dto(self, x: BanInstanceTypeRequest):
        return BanInstanceTypeRequestDTO(
            instance_type = x.instance_type.get_value(),
            region_id = x.region_id.get_value(),
            zone_id = x.zone_id.get_value(),
            duration = None if x.duration is None else x.duration.get_value()
        )
ban_instance_type_request_assembler = BanInstanceTypeRequestAssembler()

class BannedInstanceTypeResponseAssembler(Assembler):
    def to_entity(self, dto: BannedInstanceTypeResponseDTO):
        return BannedInstanceTypeResponse(
            instance_types = [InstanceTypeID(m) for m in dto.instance_types],
            region_id = RegionID(dto.region_id),
            zone_id = None if dto.zone_id is None else ZoneID(dto.zone_id)
        )

    def to_dto(self, x: BannedInstanceTypeResponse):
        return BannedInstanceTypeResponseDTO(
            instance_types = [m.get_value() for m in x.instance_types],
            region_id = x.region_id.get_value(),
            zone_id = None if x.zone_id is None else x.zone_id.get_value()
        )
banned_instance_type_response_assembler = BannedInstanceTypeResponseAssembler()

class CommandSettingDetailAssembler(Assembler):
    def to_entity(self, dto: CommandSettingDetailDTO):
        return CommandSettingDetail(
            command=Command(dto.command),
            module=CommandModule(dto.module),
            retries=Number(dto.retries),
            timeout=Number(dto.timeout),
            delay=Number(dto.delay)
        )
    def to_dto(self, x: CommandSettingDetail):
        return CommandSettingDetailDTO(
            command=x.command.get_value(),
            module=x.module.get_value(),
            retries=x.retries.get_value(),
            timeout=x.timeout.get_value(),
            delay=x.delay.get_value()
        )
command_setting_detail_assembler = CommandSettingDetailAssembler()
class CommandHostAssembler(Assembler):
    def to_entity(self, dto: CommandHostDTO):
        return CommandHost(
            ip=IP(dto.ip),
            hostname=None if dto.hostname is None else Name(dto.hostname),
            port=Number(dto.port),
            username=Name(dto.username)
        )
    def to_dto(self, x: CommandHost):
        return CommandHostDTO(
            ip=x.ip.get_value(),
            hostname=None if x.hostname is None else x.hostname.get_value(),
            port=x.port.get_value(),
            username=x.username.get_value(),
        )
command_host_assembler = CommandHostAssembler()

class CommandContextAssembler(Assembler):
    def to_entity(self, dto: CommandContextDTO):
        return CommandContext(
            connection=U(dto.connection),
            module_path=Path(dto.module_path),
            forks=Number(dto.forks),
            timeout=Number(dto.timeout),
            remote_user=Name(dto.remote_user),
            ask_pass=Bool(dto.ask_pass),
            ssh_extra_args=None if dto.ssh_extra_args is None else U(dto.ssh_extra_args),
            sftp_extra_args=None if dto.sftp_extra_args is None else U(dto.sftp_extra_args),
            scp_extra_args=None if dto.scp_extra_args is None else U(dto.scp_extra_args),
            ask_value_pass=Bool(dto.ask_value_pass),
            listhosts=Bool(dto.listhosts),
            listtasks=Bool(dto.listtasks),
            listtags=Bool(dto.listtags),
            syntax=Bool(dto.syntax),
            ssh_common_args=U(dto.ssh_common_args),
            become=None if dto.become is None else U(dto.become),
            become_method=None if dto.become_method is None else U(dto.become_method),
            become_user=Name(dto.become_user),
            check=Bool(dto.check),
            diff=Bool(dto.diff),
            verbosity=Number(dto.verbosity)
        )
    def to_dto(self, x: CommandContext):
        return CommandContextDTO(
            connection=x.connection.get_value(),
            module_path=x.module_path.get_value(),
            forks=x.forks.get_value(),
            timeout=x.timeout.get_value(),
            remote_user=x.remote_user.get_value(),
            ask_pass=x.ask_pass.get_value(),
            ssh_extra_args=None if x.ssh_extra_args is None else x.ssh_extra_args.get_value(),
            sftp_extra_args=None if x.sftp_extra_args is None else x.sftp_extra_args.get_value(),
            scp_extra_args=None if x.scp_extra_args is None else x.scp_extra_args.get_value(),
            ask_value_pass=x.ask_value_pass.get_value(),
            listhosts=x.listhosts.get_value(),
            listtasks=x.listtasks.get_value(),
            listtags=x.listtags.get_value(),
            syntax=x.syntax.get_value(),
            ssh_common_args=x.ssh_common_args.get_value(),
            become=None if x.become is None else x.become.get_value(),
            become_method=None if x.become_method is None else x.become_method.get_value(),
            become_user=x.become_user.get_value(),
            check=x.check.get_value(),
            diff=x.diff.get_value(),
            verbosity=x.verbosity.get_value()
        )
command_context_assembler = CommandContextAssembler()

class CommandRequestAssembler(Assembler):
    def to_entity(self, dto: CommandRequestDTO):
        return CommandRequest(
            commands=[command_setting_detail_assembler.to_entity(m) for m in dto.commands],
            hosts = [command_host_assembler.to_entity(m) for m in dto.hosts],
            context=None if dto.context is None else command_context_assembler.to_entity(dto.context),
            priority=CommandPriority(dto.priority),
            timeout=Number(dto.timeout),
        )
    def to_dto(self, x: CommandRequest):
        return CommandRequestDTO(
            commands = [command_setting_detail_assembler.to_dto(m) for m in x.commands],
            hosts=[command_host_assembler.to_dto(m) for m in x.hosts],
            context=None if x.context is None else command_context_assembler.to_dto(x.context),
            priority=x.priority.get_value(),
            timeout=x.timeout.get_value(),
        )
command_request_assembler = CommandRequestAssembler()

class CommandResponseAssembler(Assembler):
    def to_entity(self, dto: CommandResponseDTO):
        return CommandResponse(
            status=CommandResponseStatus(dto.status),
            ip=IP(dto.ip),
            message=Text(dto.message),
            exception=None if dto.exception is None else Text(dto.exception),
            stdout=None if dto.stdout is None else Text(dto.stdout),
            cmd=None if dto.cmd is None else Text(dto.cmd),
            start_time=None if dto.start_time is None else DateTime(dto.start_time),
            end_time = None if dto.end_time is None else DateTime(dto.end_time),
            delta_time = None if dto.delta_time is None else DeltaTime(dto.delta_time),
            stderr=None if dto.stderr is None else Text(dto.stderr),
            entry_time=None if dto.entry_time is None else DateTime(dto.entry_time),
            exit_time=None if dto.exit_time is None else DateTime(dto.exit_time)
        )
    def to_dto(self, x: CommandResponse):
        return CommandResponseDTO(
            status=x.status.get_value(),
            ip=x.ip.get_value(),
            message=x.message.get_value(),
            exception=None if x.exception is None else x.exception.get_value(),
            stdout=None if x.stdout is None else x.stdout.get_value(),
            cmd=None if x.cmd is None else x.cmd.get_value(),
            start_time=None if x.start_time is None else x.start_time.get_value(),
            end_time=None if x.end_time is None else x.end_time.get_value(),
            delta_time=None if x.delta_time is None else x.delta_time.get_value(),
            stderr=None if x.stderr is None else x.stderr.get_value(),
            entry_time=None if x.entry_time is None else x.entry_time.get_value(),
            exit_time=None if x.exit_time is None else x.exit_time.get_value()
        )
command_response_assembler = CommandResponseAssembler()

class CommandItemAssembler(Assembler):
    def to_entity(self, dto: CommandItemDTO):
        return CommandItem(
            id=InstanceCreationID(dto.id),
            command_request=None if dto.command_request is None else command_request_assembler.to_entity(dto.command_request),
            creation_time=CreationTime(dto.creation_time),
            status=CommandStatus(dto.status),
            details=None if dto.details is None else [command_response_assembler.to_entity(m) for m in dto.details],
            exception=None if dto.exception is None else CommandException(dto.exception),
            _life_time = Number(dto._life_time)
        )
    def to_dto(self, x: CommandItem):
        return CommandItemDTO(
            id=x.id.get_value(),
            command_request=None if x.command_request is None else command_request_assembler.to_dto(x.command_request),
            creation_time = x.creation_time.get_value().replace(microsecond=0).isoformat(),
            status=x.status.get_value(),
            details=None if x.details is None else [command_response_assembler.to_dto(m) for m in x.details],
            exception=None if x.exception is None else x.exception.get_value(),
            _life_time=x._life_time.get_value()
        )
command_item_assembler = CommandItemAssembler()

class DNSRecordAssembler(Assembler):
    def to_entity(self, dto: DNSRecordDTO):
        return DNSRecord(
            domain_name = DomainName(dto.domain_name),
            subdomain = Subdomain(dto.subdomain),
            value = Value(dto.value),
            id = ID(dto.id),
            weight = Weight(dto.weight),
            dns_type = DNSType(dto.dns_type),
            ttl = Number(dto.ttl),
            priority = Priority(dto.priority),
            line = Line(dto.line)
        )
    def to_dto(self, x: DNSRecord):
        return DNSRecordDTO(
            domain_name = None if x.domain_name is None else x.domain_name.get_value(),
            subdomain = None if x.subdomain is None else x.subdomain.get_value(),
            value = None if x.value is None else x.value.get_value(),
            id = None if x.id is None else x.id.get_value(),
            weight = None if x.weight is None else x.weight.get_value(),
            dns_type = None if x.dns_type is None else x.dns_type.get_value(),
            ttl = None if x.ttl is None else x.ttl.get_value(),
            priority = None if x.priority is None else x.priority.get_value(),
            line = None if x.line is None else x.line.get_value()
        )



from ddd_objects.domain.value_obj import ValueObject

class PipelineIndicator(ValueObject):
    def __init__(self, value='run') -> None:
        self.value = value

    def is_wait(self):
        return self.value=='wait'
    
    def is_pass(self):
        return self.value=='pass'

    def is_run(self):
        return self.value=='run'
    
    def is_run_again(self):
        return self.value=='run again'

    def is_over(self):
        return self.value=='over'
    
    def set_wait(self):
        self.value='wait'
        return self
    
    def set_pass(self):
        self.value='pass'
        return self

    def set_run(self):
        self.value='run'
        return self

    def set_run_again(self):
        self.value='run again'
        return self

    def set_over(self):
        self.value='over'
        return self

    def __str__(self) -> str:
        return self.value
    
    def __repr__(self) -> str:
        return self.value

class PipelineContext(ValueObject):
    def __repr__(self) -> str:
        return str(self.value)

class Function(ValueObject):
    pass

class CommandRequestID(ValueObject):
    pass

class PipelineOutput(ValueObject):
    pass


from typing import List
from ddd_objects.infrastructure.converter import Converter

from ..domain.entity import (
    BanInstanceTypeRequest,
    BannedInstanceTypeResponse,
    CommandContext,
    CommandHost,
    CommandItem,
    CommandRequest,
    CommandResponse,
    CommandSettingDetail,
    Condition,
    InstanceCreationItem,
    InstanceCreationRequest,
    InstanceUserSetting,
    InstanceInfo,
    DNSRecord,
    InstanceSetting,
)
from ..domain.value_obj import (
    CommandException,
    InstanceException,
    InstanceName,
    KeyName,
    U,
    Command,
    CommandID,
    CommandModule,
    CommandPriority,
    CommandResponseStatus,
    CommandStatus,
    CreationTime,
    Data,
    DeltaTime,
    InstanceCreationID,
    InstanceCreationPriority,
    InstanceCreationStatus,
    MaxCPUNumber,
    MaxGPUMemorySize,
    MaxGPUNumber,
    MaxMemorySize,
    MinCPUNumber,
    MinGPUMemorySize,
    MinGPUNumber,
    MinMemorySize,
    Password,
    Path,
    Size,
    Type,
    IP,
    Bandwidth,
    SecurityGroupID,
    InstanceInfoID,
    ImageID,
    DNSType,
    Price,
    Bool,
    RegionID,
    ID,
    InstanceTypeID,
    Priority,
    InstanceStatus,
    Text,
    DomainName,
    Line,
    Value,
    Number,
    Name,
    Subdomain,
    Weight,
    DateTime,
    ZoneID
)
from .do import (
    BanInstanceTypeRequestDO,
    BannedInstanceTypeResponseDO,
    CommandContextDO,
    CommandHostDO,
    CommandResponseDO,
    CommandSettingDetailDO,
    InstanceInfoDO,
    DNSRecordDO,
    InstanceSettingDO,
    ConditionDO,
    InstanceUserSettingDO,
    InstanceCreationRequestDO,
    InstanceCreationItemDO,
    CommandRequestDO,
    CommandItemDO
)

class ConditionConverter(Converter):
    def to_entity(self, do: ConditionDO):
        return Condition(
            min_cpu_num=MinCPUNumber(do.min_cpu_num),
            max_cpu_num=MaxCPUNumber(do.max_cpu_num),
            min_memory_size=MinMemorySize(do.min_memory_size),
            max_memory_size=MaxMemorySize(do.max_memory_size),
            min_gpu_num=None if do.min_gpu_num is None else MinGPUNumber(do.min_gpu_num),
            max_gpu_num=None if do.max_gpu_num is None else MaxGPUNumber(do.max_gpu_num),
            min_gpu_memory_size=None if do.min_gpu_memory_size is None else MinGPUMemorySize(do.min_gpu_memory_size),
            max_gpu_memory_size=None if do.max_gpu_memory_size is None else MaxGPUMemorySize(do.max_gpu_memory_size)
        )
    
    def to_do(self, x: Condition):
        return ConditionDO(
            min_cpu_num=x.min_cpu_num.get_value(),
            max_cpu_num=x.max_cpu_num.get_value(),
            min_memory_size=x.min_memory_size.get_value(),
            max_memory_size=x.max_memory_size.get_value(),
            min_gpu_num=None if x.min_gpu_num is None else x.min_gpu_num.get_value(),
            max_gpu_num=None if x.max_gpu_num is None else x.max_gpu_num.get_value(),
            min_gpu_memory_size=None if x.min_gpu_memory_size is None else x.min_gpu_memory_size.get_value(),
            max_gpu_memory_size=None if x.max_gpu_memory_size is None else x.max_gpu_memory_size.get_value()
        )
condition_converter = ConditionConverter()

class InstanceUserSettingConverter(Converter):
    def to_entity(self, do: InstanceUserSettingDO):
        return InstanceUserSetting(
            name = InstanceName(do.name),
            password = None if do.password is None else Password(do.password),
            amount=Number(do.amount),
            image_id=None if do.image_id is None else ImageID(do.image_id),
            region_id=RegionID(do.region_id),
            internet_pay_type=None if do.internet_pay_type is None else Type(do.internet_pay_type),
            bandwidth_in=Bandwidth(do.bandwidth_in),
            bandwidth_out=Bandwidth(do.bandwidth_out),
            user_data=None if do.user_data is None else Data(do.user_data),
            disk_size=Size(do.disk_size),
            key_name=KeyName(do.key_name),
            exclude_instance_types=[InstanceTypeID(m) for m in do.exclude_instance_types],
            inner_connection=Bool(do.inner_connection)
        )
    def to_do(self, x:InstanceUserSetting):
        return InstanceUserSettingDO(
            name=x.name.get_value(),
            password=None if x.password is None else x.password.get_value(),
            amount=x.amount.get_value(),
            image_id=None if x.image_id is None else x.image_id.get_value(),
            region_id=x.region_id.get_value(),
            internet_pay_type=None if x.internet_pay_type is None else x.internet_pay_type.get_value(),
            bandwidth_in=x.bandwidth_in.get_value(),
            bandwidth_out=x.bandwidth_out.get_value(),
            user_data=None if x.user_data is None else x.user_data.get_value(),
            disk_size=x.disk_size.get_value(),
            key_name=x.key_name.get_value(),
            exclude_instance_types=[m.get_value() for m in x.exclude_instance_types],
            inner_connection=x.inner_connection.get_value()
        )
instance_user_setting_converter = InstanceUserSettingConverter()

class InstanceCreationRequestConverter(Converter):
    def to_entity(self, do: InstanceCreationRequestDO):
        return InstanceCreationRequest(
            instance_user_setting=instance_user_setting_converter.to_entity(do.instance_user_setting),
            condition=condition_converter.to_entity(do.condition),
            priority=InstanceCreationPriority(do.priority),
            timeout=Number(do.timeout)
        )
    def to_do(self, x: InstanceCreationRequest):
        return InstanceCreationRequestDO(
            instance_user_setting=instance_user_setting_converter.to_do(x.instance_user_setting),
            condition=condition_converter.to_do(x.condition),
            priority=x.priority.get_value(),
            timeout=x.timeout.get_value()
        )
instance_creation_request_converter = InstanceCreationRequestConverter()

class InstanceCreationItemConverter(Converter):
    def to_entity(self, do: InstanceCreationItemDO):
        return InstanceCreationItem(
            id=InstanceCreationID(do.id),
            instance_creation_request=None if do.instance_creation_request is None else instance_creation_request_converter.to_entity(do.instance_creation_request),
            creation_time=CreationTime(do.creation_time),
            status=InstanceCreationStatus(do.status),
            details=None if do.details is None else [instance_info_converter.to_entity(m) for m in do.details],
            entry_time=None if do.entry_time is None else DateTime(do.entry_time),
            exit_time=None if do.exit_time is None else DateTime(do.exit_time),
            exception=None if do.exception is None else InstanceException(do.exception),
            _life_time=Number(do._life_time)
        )
    def to_do(self, x: InstanceCreationItem):
        return InstanceCreationItemDO(
            id = x.id.get_value(),
            instance_creation_request = None if x.instance_creation_request is None else instance_creation_request_converter.to_do(x.instance_creation_request),
            creation_time=x.creation_time.get_value().replace(microsecond=0).isoformat(),
            status=x.status.get_value(),
            details=None if x.details is None else [instance_info_converter.to_do(m) for m in x.details],
            entry_time=None if x.entry_time is None else x.entry_time.get_value(),
            exit_time=None if x.exit_time is None else x.exit_time.get_value(),
            exception=None if x.exception is None else x.exception.get_value(),
            _life_time=x._life_time.get_value()
        )
instance_creation_item_converter = InstanceCreationItemConverter()

class BanInstanceTypeRequestConverter(Converter):
    def to_entity(self, do: BanInstanceTypeRequestDO):
        return BanInstanceTypeRequest(
            instance_type = InstanceTypeID(do.instance_type),
            region_id = RegionID(do.region_id),
            zone_id = ZoneID(do.zone_id),
            duration = None if do.duration is None else Number(do.duration)
        )
    def to_do(self, x: BanInstanceTypeRequest):
        return BanInstanceTypeRequestDO(
            instance_type = x.instance_type.get_value(),
            region_id = x.region_id.get_value(),
            zone_id = x.zone_id.get_value(),
            duration = None if x.duration is None else x.duration.get_value()
        )
ban_instance_type_request_converter = BanInstanceTypeRequestConverter()


class BannedInstanceTypeResponseConverter(Converter):
    def to_entity(self, do: BannedInstanceTypeResponseDO):
        return BannedInstanceTypeResponse(
            instance_types=[InstanceTypeID(m) for m in do.instance_types],
            region_id = RegionID(do.region_id),
            zone_id = ZoneID(do.zone_id)
        )
    
    def to_do(self, x: BannedInstanceTypeResponse):
        return BannedInstanceTypeResponseDO(
            instance_types=[m.get_value() for m in x.instance_types],
            region_id = x.region_id.get_value(),
            zone_id = x.zone_id.get_value()
        )
banned_instance_type_response_converter = BannedInstanceTypeResponseConverter()


class CommandSettingDetailConverter(Converter):
    def to_entity(self, do: CommandSettingDetailDO):
        return CommandSettingDetail(
            command=Command(do.command),
            module=CommandModule(do.module),
            retries=Number(do.retries),
            timeout=Number(do.timeout),
            delay=Number(do.delay)
        )
    def to_do(self, x: CommandSettingDetail):
        return CommandSettingDetailDO(
            command=x.command.get_value(),
            module=x.module.get_value(),
            retries=x.retries.get_value(),
            timeout=x.timeout.get_value(),
            delay=x.delay.get_value()
        )
command_setting_detail_converter = CommandSettingDetailConverter()

class CommandHostConverter(Converter):
    def to_entity(self, do: CommandHostDO):
        return CommandHost(
            ip=IP(do.ip),
            hostname=None if do.hostname is None else Name(do.hostname),
            port=Number(do.port),
            username=Name(do.username)
        )
    def to_do(self, x: CommandHost):
        return CommandHostDO(
            ip=x.ip.get_value(),
            hostname=None if x.hostname is None else x.hostname.get_value(),
            port=x.port.get_value(),
            username=x.username.get_value()
        )
command_host_converter = CommandHostConverter()

class CommandContextConverter(Converter):
    def to_entity(self, do: CommandContextDO):
        return CommandContext(
            connection=U(do.connection),
            module_path=Path(do.module_path),
            forks=Number(do.forks),
            timeout=Number(do.timeout),
            remote_user=Name(do.remote_user),
            ask_pass=Bool(do.ask_pass),
            ssh_extra_args=None if do.ssh_extra_args is None else U(do.ssh_extra_args),
            sftp_extra_args=None if do.sftp_extra_args is None else U(do.sftp_extra_args),
            scp_extra_args=None if do.scp_extra_args is None else U(do.scp_extra_args),
            ask_value_pass=Bool(do.ask_value_pass),
            listhosts=Bool(do.listhosts),
            listtasks=Bool(do.listtasks),
            listtags=Bool(do.listtags),
            syntax=Bool(do.syntax),
            ssh_common_args=U(do.ssh_common_args),
            become=None if do.become is None else U(do.become),
            become_method=None if do.become_method is None else U(do.become_method),
            become_user=Name(do.become_user),
            check=Bool(do.check),
            diff=Bool(do.diff),
            verbosity=Number(do.verbosity)
        )
    def to_do(self, x: CommandContext):
        return CommandContextDO(
            connection=x.connection.get_value(),
            module_path=x.module_path.get_value(),
            forks=x.forks.get_value(),
            timeout=x.timeout.get_value(),
            remote_user=x.remote_user.get_value(),
            ask_pass=x.ask_pass.get_value(),
            ssh_extra_args=None if x.ssh_extra_args is None else x.ssh_extra_args.get_value(),
            sftp_extra_args=None if x.sftp_extra_args is None else x.sftp_extra_args.get_value(),
            scp_extra_args=None if x.scp_extra_args is None else x.scp_extra_args.get_value(),
            ask_value_pass=x.ask_value_pass.get_value(),
            listhosts=x.listhosts.get_value(),
            listtasks=x.listtasks.get_value(),
            listtags=x.listtags.get_value(),
            syntax=x.syntax.get_value(),
            ssh_common_args=x.ssh_common_args.get_value(),
            become=None if x.become is None else x.become.get_value(),
            become_method=None if x.become_method is None else x.become_method.get_value(),
            become_user=x.become_user.get_value(),
            check=x.check.get_value(),
            diff=x.diff.get_value(),
            verbosity=x.verbosity.get_value()

        )
command_context_converter = CommandContextConverter()

class CommandRequestConverter(Converter):
    def to_entity(self, do: CommandRequestDO):
        return CommandRequest(
            commands=[command_setting_detail_converter.to_entity(m) for m in do.commands],
            hosts=[command_host_converter.to_entity(m) for m in do.hosts],
            context=None if do.context is None else command_context_converter.to_entity(do.context),
            priority=CommandPriority(do.priority),
            timeout=Number(do.timeout),
        )
    
    def to_do(self, x: CommandRequest):
        return CommandRequestDO(
            commands=[command_setting_detail_converter.to_do(m) for m in x.commands],
            hosts=[command_host_converter.to_do(m) for m in x.hosts],
            context=None if x.context is None else command_context_converter.to_do(x.context),
            priority=x.priority.get_value(),
            timeout=x.timeout.get_value(),
        )
command_request_converter = CommandRequestConverter()

class CommandResponseConverter(Converter):
    def to_entity(self, do: CommandResponseDO):
        return CommandResponse(
            status=CommandResponseStatus(do.status),
            ip=IP(do.ip),
            message=Text(do.message),
            exception=None if do.exception is None else Text(do.exception),
            stdout=None if do.stdout is None else Text(do.stdout),
            cmd=None if do.cmd is None else Text(do.cmd),
            start_time=None if do.start_time is None else DateTime(do.start_time),
            end_time = None if do.end_time is None else DateTime(do.end_time),
            delta_time = None if do.delta_time is None else DeltaTime(do.delta_time),
            stderr=None if do.stderr is None else Text(do.stderr),
            entry_time=None if do.entry_time is None else DateTime(do.entry_time),
            exit_time=None if do.exit_time is None else DateTime(do.exit_time)
        )
    def to_do(self, x: CommandResponse):
        return CommandResponseDO(
            status=x.status.get_value(),
            ip=x.ip.get_value(),
            message=x.message.get_value(),
            exception=None if x.exception is None else x.exception.get_value(),
            stdout=None if x.stdout is None else x.stdout.get_value(),
            cmd=None if x.cmd is None else x.cmd.get_value(),
            start_time=None if x.start_time is None else x.start_time.get_value(),
            end_time=None if x.end_time is None else x.end_time.get_value(),
            delta_time=None if x.delta_time is None else x.delta_time.get_value(),
            stderr=None if x.stderr is None else x.stderr.get_value(),
            entry_time=None if x.entry_time is None else x.entry_time.get_value(),
            exit_time=None if x.exit_time is None else x.exit_time.get_value()
        )
command_response_converter = CommandResponseConverter()

class CommandItemConverter(Converter):
    def to_entity(self, do: CommandItemDO):
        return CommandItem(
            id=CommandID(do.id),
            command_request=None if do.command_request is None else command_request_converter.to_entity(do.command_request),
            creation_time=CreationTime(do.creation_time),
            status=CommandStatus(do.status),
            details=None if do.details is None else [command_response_converter.to_entity(m) for m in do.details],
            exception=None if do.exception is None else CommandException(do.exception),
            _life_time=Number(do._life_time)
        )
    def to_do(self, x: CommandItem):
        return CommandItemDO(
            id=x.id.get_value(),
            command_request=None if x.command_request is None else command_request_converter.to_do(x.command_request),
            creation_time=x.creation_time.get_value().replace(microsecond=0).isoformat(),
            status=x.status.get_value(),
            details=None if x.details is None else [command_response_converter.to_do(m) for m in x.details],
            exception=None if x.exception is None else x.exception.get_value(),
            _life_time=x._life_time.get_value()
        )
command_item_converter = CommandItemConverter()

class InstanceInfoConverter(Converter):
    @Converter.convert_none
    def to_entity(self, do: InstanceInfoDO):
        return InstanceInfo(
            id = InstanceInfoID(do.id),
            instance_type=InstanceTypeID(do.instance_type),
            create_time=DateTime(do.create_time),
            name=Name(do.name),
            hostname=Name(do.hostname),
            pay_type=Type(do.pay_type),
            public_ip=[IP(ip) for ip in do.public_ip],
            private_ip=None if do.private_ip is None else IP(do.private_ip),
            os_name=Name(do.os_name),
            price=Price(do.price),
            image_id=ImageID(do.image_id),
            region_id=RegionID(do.region_id),
            zone_id=ZoneID(do.zone_id),
            internet_pay_type=Type(do.internet_pay_type),
            bandwidth_in=Bandwidth(do.bandwidth_in),
            bandwidth_out=Bandwidth(do.bandwidth_out),
            security_group_id=[SecurityGroupID(id) for id in do.security_group_id],
            instance_expired_time=DateTime(do.instance_expired_time),
            auto_release_time=DateTime(do.auto_release_time),
            status = InstanceStatus(do.status),
            key_name = KeyName(do.key_name)
        )
    def to_do(self, x: InstanceInfo):
        return InstanceInfoDO(
            id = x.id.get_value(),
            instance_type = x.instance_type.get_value(),
            create_time = x.create_time.get_value(),
            name = x.name.get_value(),
            hostname = x.name.get_value(),
            pay_type = x.pay_type.get_value(),
            public_ip = [m.get_value() for m in x.public_ip],
            private_ip = None if x.private_ip is None else x.private_ip.get_value(),
            os_name = x.os_name.get_value(),
            price = x.price.get_value(),
            image_id = x.image_id.get_value(),
            region_id = x.region_id.get_value(),
            zone_id = x.zone_id.get_value(),
            internet_pay_type=x.internet_pay_type.get_value(),
            bandwidth_in = x.bandwidth_in.get_value(),
            bandwidth_out = x.bandwidth_out.get_value(),
            security_group_id = [m.get_value() for m in x.security_group_id],
            instance_expired_time = x.instance_expired_time.get_value(),
            auto_release_time = x.auto_release_time.get_value(),
            status = x.status.get_value(),
            key_name = x.key_name.get_value()
        )
instance_info_converter = InstanceInfoConverter()

class InstanceSettingConverter(Converter):
    @Converter.convert_none
    def to_do(self, x: InstanceSetting):
        return InstanceSettingDO(
            region_id = x.region_id.get_value() ,
            zone_id = x.zone_id.get_value(),
            spot_price=x.spot_price.get_value(),
            instance_type=x.instance_type.get_value(),
            amount=x.amount.get_value(),
            image_id=None if x.image_id is None else x.image_id.get_value(),
            instance_name= x.name.get_value() ,
            password=None if x.password is None else x.password.get_value(),
            disk_size=x.disk_size.get_value(),
            bandwidth_in=x.bandwidth_in.get_value(),
            bandwidth_out=x.bandwidth_out.get_value(),
            user_data=None if x.user_data is None else x.user_data.get_value(),
            internet_pay_type=None if x.internet_pay_type is None else x.internet_pay_type.get_value(),
            security_group_id=x.security_group_id.get_value(),
            key_name=x.key_name.get_value()
        )
instance_setting_converter = InstanceSettingConverter()

class DNSRecordConverter(Converter):
    def to_entity(self, do: DNSRecordDO):
        return DNSRecord(
            domain_name = DomainName(do.domain_name),
            subdomain = Subdomain(do.subdomain),
            value = Value(do.value),
            id = ID(do.id),
            weight = Weight(do.weight),
            dns_type = DNSType(do.dns_type),
            ttl = Number(do.ttl),
            priority = Priority(do.priority),
            line = Line(do.line)
        )
    def to_do(self, x: DNSRecord):
        return DNSRecordDO(
            domain_name = None if x.domain_name is None else x.domain_name.get_value(),
            subdomain = None if x.subdomain is None else x.subdomain.get_value(),
            value = None if x.value is None else x.value.get_value(),
            id = None if x.id is None else x.id.get_value(),
            weight = None if x.weight is None else x.weight.get_value(),
            dns_type = None if x.dns_type is None else x.dns_type.get_value(),
            ttl = None if x.ttl is None else x.ttl.get_value(),
            priority = None if x.priority is None else x.priority.get_value(),
            line = None if x.line is None else x.line.get_value()
        )
dns_record_converter = DNSRecordConverter()



import os
from log_control_ao import Logger

logger = Logger(domain='vpc-ao', local=True)
vpc_token = os.environ.get('VPC_TOKEN', 'test')
vpc_ip = 'vpc-control-test-svc.system-service.svc.cluster.local'
vpc_port = 8080



import traceback, time
from retry import retry
from typing import List, Tuple, Dict, Optional, Callable
from ddd_objects.domain.exception import OperationError
from ddd_objects.lib import get_random_string
from .repository import CommandRepository
from .value_obj_ext import (
    PipelineOutput,
    PipelineIndicator,
    PipelineContext,
)
from .value_obj import Name, Command, Number, CommandPriority, CommandID
from .entity_ext import (
    PipelineActionOutput,
    _PipelineStage
)
from .entity import CommandHost, CommandRequest, CommandSettingDetail, CommandItem
from ..settings import logger as _logger



class CyclePipelineTemplate:
    def __init__(
        self, 
        stages:List[_PipelineStage], 
        context: PipelineContext,
        output_fn: Optional[Callable]=None,
    ) -> None:
        self.__stages = stages
        self.__stage_idx_map = {s.stage_name:i for i,s in enumerate(stages)}
        self.__curr_stage_idx = -1
        self.__stage_num = len(stages)
        self.__output_fn = output_fn
        self.context = context
        self.indicator = PipelineIndicator()
        self.status = {}
    

    def __next_stage(self)->Tuple[str, Callable, bool]:
        self.__curr_stage_idx = (self.__curr_stage_idx+1)%self.__stage_num
        stage = self.__stages[self.__curr_stage_idx]
        return stage.stage_name, stage.action, stage.allow_failure
    

    def __jump_to(self, stage:Name):
        stage_idx = self.__stage_idx_map[stage.get_value()]
        self.__curr_stage_idx = (stage_idx-1)%self.__stage_num


    def run(self):
        indicator = self.indicator
        if indicator.is_wait() or indicator.is_over():
            pass

        elif indicator.is_pass():
            stage, func, allow_failure = self.__next_stage()
            self.status = {
                'stage': stage,
                'action': func,
                'context': self.context,
                'indictor': self.indicator,
                'allow_failure': allow_failure
            }

        elif indicator.is_run() or indicator.is_run_again():
            if indicator.is_run():
                stage, func, allow_failure = self.__next_stage()
                self.status['stage'] = stage
                self.status['action'] = func
                self.status['allow_failure'] = allow_failure
            else:
                func = self.status['action']
                allow_failure = self.status['allow_failure']
                stage = self.status['stage']
            try:
                output:PipelineActionOutput = func(self.context)
                self.context = output.context
                self.indicator = output.indicator
                self.output = output.output
                jump_stage = output.jump_stage
                if jump_stage is not None:
                    self.__jump_to(jump_stage)
                self.status = {
                    'stage': stage,
                    'action': func,
                    'context': self.context,
                    'indictor': self.indicator,
                    'allow_failure': allow_failure,
                    'output': self.output
                }
            except:
                if not allow_failure:
                    raise OperationError(traceback.format_exc())
                else:
                    self.indicator.set_run_again()
        return self.status

    def is_over(self):
        return self.indicator.is_over()

    def get_output(self, x=None):
        if self.__output_fn is None:
            return self.output
        else:
            return self.__output_fn(self.output, x)

    def run_until_complete(self, timeout=100, interval=1):
        x = []
        for i in range(timeout):
            status = self.run()
            if self.output:
                # _logger.debug(f'pipeline status: {status}')
                x.append(self.output)
            if self.is_over():
                return self.get_output(x)
            time.sleep(interval)
        raise OperationError(f'Timeout ({timeout}s) when running loop')



class PipelineGenerator:

    def generate_pipeline(self):
        raise NotImplementedError

    def set_output(self, output: Optional[PipelineOutput]):
        raise NotImplementedError

    def _set_run_again_output(self, context):
        return PipelineActionOutput(
            PipelineContext(context),
            PipelineIndicator().set_run_again()
        )
    def _set_run_next_output(self, context):
        return PipelineActionOutput(PipelineContext(context))

    def _set_over_output(self, context, output=None):
        return PipelineActionOutput(
            PipelineContext(context),
            PipelineIndicator().set_over(),
            output=output
        )

    def _set_jump_output(self, context, action_name):
        return PipelineActionOutput(
            PipelineContext(context),
            PipelineIndicator(),
            Name(action_name)
        )

def combine_commands(
    commands: List[CommandSettingDetail],
    code: Optional[str] = None
)->Tuple[List[CommandSettingDetail], Number, str]:
    timeout = Number(sum([c.timeout.get_value() for c in commands]))
    commands = [c.command.get_value() for c in commands]
    if code is None:
        code = get_random_string(10)
    tmpfile = get_random_string(10)
    commands.extend([f'echo $?>/tmp/{code} || echo $?>/tmp/{code}', 'sleep 600', f'rm /tmp/{code} &'])
    command_str = ' && '.join(commands)
    command = f"echo '{command_str}'>{tmpfile}"
    return [
        CommandSettingDetail(Command(command)),
        CommandSettingDetail(Command(f'nohup /bin/bash {tmpfile} >/dev/null 2>&1')),
        CommandSettingDetail(Command(f'rm {tmpfile}'))
    ], timeout, code

def get_check_commands(code: str)->List[CommandSettingDetail]:
    commands = [
        CommandSettingDetail(Command(f'cat /tmp/{code}'))
    ]
    return commands


@retry(tries=2, delay=0.5)
def send_request(command_repo: CommandRepository, request: CommandRequest):
    return command_repo.send_request(request)


@retry(tries=2, delay=0.5)
def find_item(command_repo: CommandRepository, request_id: CommandID)->CommandItem:
    return command_repo.find_item(request_id)
 
    
class HeavyCommandPipelineGenerator(PipelineGenerator):
    def __init__(
        self, 
        command_repo: CommandRepository,
        hosts: List[CommandHost],
        commands: List[CommandSettingDetail],
        priority: CommandPriority=CommandPriority(3),
        check_interval: float=5,
        timeout: float=50,
        allow_failure_num: int=10,
    ) -> None:
        self.context = PipelineContext(
            {
                'command_repo': command_repo,
                'hosts': hosts,
                'commands': commands,
                'priority': priority,
                'check_interval': check_interval,
                'timeout': timeout,
                'allow_failure_num': allow_failure_num,
                'failure_num': 0
            }
        )


    def send_request(self, context:PipelineContext):
        context = context.get_value()
        if 'end_time' not in context:
            context['end_time'] = time.time()+context['timeout']
        commands:List[CommandSettingDetail] = context['commands']
        request_id = context.get('request_id', None)
        if request_id:
            return self._set_run_next_output(context)
        send_ips = context.get('send_ips', None)
        hosts = [h for h in context['hosts'] if h.ip in send_ips] \
                if send_ips is not None \
                else context['hosts']
        if hosts:
            commands, timeout, context['code'] = combine_commands(commands, context.get('code', None))
            request = CommandRequest(
                commands=commands,
                hosts=hosts,
                context=None,
                priority=context['priority'],
                timeout=Number(3)
            )
            _logger.debug(f'send_request: {request}')
            context['request_id'] = send_request(
                context['command_repo'], request
            ) if hosts else None
        return self._set_run_next_output(context)
        

    def check_response(self, context:PipelineContext):
        context = context.get_value()
        request_id = context['request_id']
        if not request_id:
            return self._set_run_next_output(context)
        item = find_item(context['command_repo'], request_id)
        _logger.debug(f'check_response: {item}')
        if item.details:
            check_progress_obj = [r for r in item.details if r.status.is_succeed()]
            context['send_ips'] = [r.ip for r in item.details if r not in check_progress_obj]
            if 'check_ips' in context:
                context['check_ips'].extend([r.ip for r in check_progress_obj])
            else:
                context['check_ips'] = [r.ip for r in check_progress_obj]
                context['request_id'] = None
        elif item.status.is_failed():
            context['failure_num'] += 1
            if context['failure_num']>context['allow_failure_num']:
                raise OperationError(f'Command item is failed: {item}')
            else:
                _logger.error(f'Command item is failed: {item}')
                context['request_id'] = None
        return self._set_run_next_output(context)


    def send_check_request(self, context:PipelineContext):
        context = context.get_value()
        request_id = context.get('check_request_id', None)
        check_ips = context.get('check_ips')
        if request_id or not check_ips:
            return self._set_run_next_output(context)
        hosts = [h for h in context['hosts'] if h.ip in check_ips]
        commands = get_check_commands(context['code'])
        request = CommandRequest(commands=commands,
                                hosts=hosts,
                                context=None,
                                priority=context['priority'],
                                timeout = Number(3))
        _logger.debug(f'send_check_request: {request}')
        context['check_request_id'] = send_request(context['command_repo'],request
                                    ) if hosts else None
        return self._set_run_next_output(context)
    

    def find_check_response(self, context:PipelineContext):
        context = context.get_value()
        request_id = context.get('check_request_id', None)
        if not request_id:
            return self._set_run_next_output(context)
        item = find_item(context['command_repo'], request_id)
        _logger.debug(f'find_check_response: {item}')
        if item.details:
            completed_obj = [r for r in item.details if r.status.is_succeed()]
            ret_resp = []
            for r in completed_obj:
                if r.stdout.get_value().strip()!='0':
                    r.status.set_failed()
                ret_resp.append(r)
            context['check_ips'] = [ip for ip in context['check_ips'] 
                            if ip not in [r.ip for r in completed_obj]]
            context['check_request_id'] = None
            item.details = ret_resp
            return PipelineActionOutput(
                context=PipelineContext(context),
                indicator=PipelineIndicator(),
                output=item
            )
        elif item.status.is_failed():
            context['failure_num'] += 1
            if context['failure_num']>context['allow_failure_num']:
                raise OperationError(f'Fail to check progress: {item.exception.get_value()}')
            else:
                _logger.error(f'Fail to check progress: {item.exception.get_value()}')
                context['check_request_id'] = None
            return self._set_run_next_output(context)
        else:
            return self._set_run_next_output(context)

    def sleep(self, context: PipelineContext):
        context = context.get_value()
        if time.time()>context['end_time']:
            raise OperationError(f'Timeout when execute command {context["commands"]}')
        if 'check_time' not in context:
            context['check_time'] = time.time()
        
        if time.time()>context['check_time']:
            if 'check_ips' not in context or 'send_ips' not in context:
                _logger.debug('check ips and send ips not in context keywords')
                context['check_time'] += context['check_interval']
                return self._set_run_next_output(context)
            elif context['check_ips']==[] and context['send_ips']==[]:
                _logger.debug('pipeline can be over')
                return self._set_over_output(context)
            elif context['send_ips']==[]:
                context['check_time'] += context['check_interval']
                return self._set_jump_output(context, 'send_check_request')
            else:
                _logger.debug(f"check ips:{context['check_ips']}, send ips: {context['send_ips']}")
                context['check_time'] += context['check_interval']
                return self._set_run_next_output(context)
        else:
            _logger.debug('pipeline is in sleep')
            return self._set_run_again_output(context)

    def get_output(self, output, x:List[CommandItem])->CommandItem:
        # x is all output from pipeline
        resp = []
        for item in x:
            resp.extend(item.details)
        item = x[-1]
        item.details = resp
        return item

    
    def generate_pipeline(self):
        stages = [
            _PipelineStage(stage_name='send_request', action=self.send_request),
            _PipelineStage(stage_name='check_response', action=self.check_response),
            _PipelineStage(stage_name='send_check_request', action=self.send_check_request),
            _PipelineStage(stage_name='find_check_response', action=self.find_check_response),
            _PipelineStage(stage_name='sleep', action=self.sleep)
        ]
        return CyclePipelineTemplate(stages, self.context, self.get_output)



import sys, datetime, string
from ipaddress import ip_address
from dateutil import parser
from ddd_objects.domain.value_obj import ValueObject, ID, ExpiredValueObject
from ddd_objects.domain.exception import ParameterError
python_version_minor = sys.version_info.minor

class U(ValueObject):
    pass

class CreationTime(ValueObject):
    def __init__(self, value=None):
        if value is None:
            value = datetime.datetime.utcnow()
        elif isinstance(value, str):
            value = parser.parse(value)
        super().__init__(value)
    def __str__(self) -> str:
        return self.value.replace(microsecond=0).isoformat()

class Name(ValueObject):
    pass

class InstanceName(Name):
    def __init__(self, value) -> None:
        if not isinstance(value, str):
            raise ParameterError('Instance name must be string!')
        if not value:
            raise ParameterError('Instance name cannot be empty!')
        if value[0] not in string.ascii_letters:
            raise ParameterError('First letter of the instance name must be letter')
        super().__init__(value)

class KeyName(Name):
    def __init__(self, value) -> None:
        if value not in ['ansible', 'ansible-test']:
            raise ValueError('Invalid value for RSA key name')
        super().__init__(value)

class InstanceCreationID(ID):
    pass

class CommandID(ID):
    pass

class ZoneID(ID):
    pass

class InstanceTypeID(ID):
    pass

class ZoneName(Name):
    pass

class RegionID(ExpiredValueObject):
    def __init__(self, value=None):
        super().__init__(value, None)

class RegionName(Name):
    pass

class Endpoint(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Number(ValueObject):
    pass

class CPUNumber(Number):
    pass
        
GPUNumber = MinCPUNumber = MinGPUNumber = CPUNumber

class MaxCPUNumber(Number):
    pass
MaxGPUNumber = MaxCPUNumber

class Data(ValueObject):
    pass

class Size(ValueObject):
    pass

class MemorySize(ValueObject):
    pass
    
GPUMemorySize = MinGPUMemorySize = MinMemorySize = MemorySize

class MaxMemorySize(MemorySize):
    pass
MaxGPUMemorySize = MaxMemorySize

class GPUType(ValueObject):
    pass

class Price(ValueObject):
    pass

class DateTime(ValueObject):
    pass

class DeltaTime(ValueObject):
    pass

class Status(ValueObject):
    pass

class InstanceCreationStatus(ValueObject):
    def __init__(self, value=None) -> None:
        value = value or 'Pending'
        self.value = value
        self._set_changed_time()

    def _set_changed_time(self):
        self.changed_time = datetime.datetime.utcnow()

    def set_sent(self):
        self.value = 'Sent'
        self._set_changed_time()

    def set_created(self):
        self.value = 'Created'
        self._set_changed_time()

    def set_deleted(self):
        self.value = 'Deleted'
        self._set_changed_time()

    def set_succeed(self):
        self.value = 'Succeed'
        self._set_changed_time()

    def set_failed(self):
        self.value = 'Failed'
        self._set_changed_time()
    
    def is_succeed(self):
        return self.value=='Succeed'
    
    def is_failed(self):
        return self.value=='Failed'

    def is_sent(self):
        return self.value=='Sent'

    def is_created(self):
        return self.value=='Created'

    def is_pending(self):
        return self.value=='Pending'

    def is_deleted(self):
        return self.value=='Deleted'

class CommandStatus(ValueObject):
    def __init__(self, value=None) -> None:
        value = value or 'Pending'
        self.value = value
        self._set_changed_time()

    def _set_changed_time(self):
        self.changed_time = datetime.datetime.utcnow()

    def set_sent(self):
        self.value = 'Sent'
        self._set_changed_time()

    def set_created(self):
        self.value = 'Created'
        self._set_changed_time()

    def set_deleted(self):
        self.value = 'Deleted'
        self._set_changed_time()

    def set_succeed(self):
        self.value = 'Succeed'
        self._set_changed_time()

    def set_failed(self):
        self.value = 'Failed'
        self._set_changed_time()

    def is_sent(self):
        return self.value=='Sent'

    def is_created(self):
        return self.value=='Created'

    def is_pending(self):
        return self.value=='Pending'

    def is_deleted(self):
        return self.value=='Deleted'
    
    def is_succeed(self):
        return self.value=='Succeed'

    def is_failed(self):
        return self.value=='Failed'

class SecurityGroupID(ExpiredValueObject):
    def __init__(self, value=None):
        if not isinstance(value, str):
            raise ParameterError(f'ID should be a string, but got {type(value)}')
        super().__init__(value, None)

class InstanceInfoID(ID):
    pass

class Type(ValueObject):
    pass

class IP(ValueObject):
    def __init__(self, value) -> None:
        value = value.strip()
        if not self.is_ip(value):
            raise ParameterError('Input parameter is not a ip')
        self.value = value

    def is_private(self, ip)->bool:
        return ip_address(ip).is_private

    def is_ip(self, ip)->bool:
        try:
            ip_address(ip)
            return True
        except:
            return False
        

class ImageID(ID):
    pass

class Bandwidth(ValueObject):
    pass

class InstanceStatus(ValueObject):
    def is_running(self):
        return self.value=='Running'

class Password(ValueObject):
    def __init__(self, value) -> None:
        if len(value)<8:
            raise ParameterError('Password should be longer than 7')
        elif len(value)>30:
            raise ParameterError('Password should be shorter than 31')
        n_uletter, n_lletter, n_number =0, 0, 0
        for c in value:
            if c.islower():
                n_lletter += 1
            elif c.isupper():
                n_uletter += 1
            elif c.isdigit():
                n_number += 1
        if n_uletter==0 or n_lletter==0 or n_number==0:
            raise ParameterError('Password should contain a lower letter, a upper letter and a number!')
        super().__init__(value)

class InstanceID(ID):
    pass

class Key(ValueObject):
    pass

class Path(ValueObject):
    pass

class Bool(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Output(ValueObject):
    pass

class DomainName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Subdomain(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Value(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DnsType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Weight(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Priority(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class CommandPriority(ValueObject):
    def __init__(self, value):
        if not isinstance(value, int):
            raise ParameterError(f'Value of CommandPriority should be int, not {type(value)}')
        if value<1 or value>5:
            raise ParameterError(f'Value of CommandPrority should be greater than 0 and less than 6')
        super().__init__(value)

    def __gt__(self, other):
        return self.value>other.value

    def __lt__(self, other):
        return self.value<other.value

    def __eq__(self, other):
        return self.value==other.value

class InstanceCreationPriority(ValueObject):
    def __init__(self, value):
        if not isinstance(value, int):
            raise ParameterError(f'Value of InstanceCreationPriority should be int, not {type(value)}')
        if value<1 or value>5:
            raise ParameterError(f'Value of InstanceCreationPrority should be greater than 0 and less than 6')
        super().__init__(value)
    
    def __gt__(self, other):
        return self.value>other.value

    def __lt__(self, other):
        return self.value<other.value

    def __eq__(self, other):
        return self.value==other.value

class Line(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class DNSType(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class ID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Port(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class String(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Text(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class BucketName(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class VersionID(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class VersionCreationTime(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class Command(ExpiredValueObject):
    def __init__(self, value):
        super().__init__(value, None)

class CommandModule(ValueObject):
    pass

class CommandResponseStatus(ValueObject):
    def is_succeed(self)->bool:
        return self.value=='succeed'
    def is_failed(self)->bool:
        return self.value=='failed'
    def is_unreachable(self)->bool:
        return self.value=='unreachable'
    def set_succeed(self):
        self.value='succeed'
    def set_failed(self):
        self.value='failed'
    def set_unreachable(self):
        self.value='unreachable'

class InstanceException(ValueObject):
    pass

class CommandException(ValueObject):
    pass




from typing import Dict
from ddd_objects.domain.repository import Repository
from .entity import CommandRequest, CommandItem
from .value_obj import CommandID

class CommandRepository(Repository):
    def get_queue_info(self)->Dict:
        raise NotImplementedError

    def send_request(self, request:CommandRequest)->CommandID:
        raise NotImplementedError
    
    def find_item(self, request_id: CommandID)->CommandItem:
        raise NotImplementedError
    
    def delete_item(self, request_id: CommandID)->bool:
        raise NotImplementedError


from ddd_objects.infrastructure.ao import exception_class_dec
from ddd_objects.infrastructure.repository_impl import error_factory
from ddd_objects.domain.exception import return_codes
import json, requests
from typing import List, Optional
from .do import (
    ReleaseInstanceInfoDO,
    CommandContextDO,
    CommandSettingDetailDO,
    CommandHostDO,
    BanInstanceTypeRequestDO,
    BannedInstanceTypeResponseDO,
    CommandItemDO,
    CommandRequestDO,
    ConditionDO,
    DNSRecordDO,
    InstanceCreationItemDO,
    InstanceCreationRequestDO,
    InstanceInfoDO,
    InstanceUserSettingDO,
    CommandResponseDO,
)

class VPCController:
    def __init__(self, ip:str, port:str, token:str) -> None:
        self.url = f"http://{ip}:{port}"
        self.header = {"api-token":token}

    def _check_error(self, status_code, info):
        if status_code>299:
            if isinstance(info['detail'], str):
                return_code = return_codes['OTHER_CODE']
                error_traceback = info['detail']
            else:
                return_code = info['detail']['return_code']
                error_traceback = info['detail']['error_traceback']
            raise error_factory.make(return_code)(error_traceback)

    def check_connection(self, timeout=3):
        response=requests.get(f'{self.url}', headers=self.header, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        if info['message']=='Hello World':
            return True
        else:
            return False

    def get_vpc_info(self, timeout=3):
        response=requests.get(f'{self.url}/vpc/info', headers=self.header, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return info

    def new_instance(self, condition:ConditionDO, setting:InstanceUserSettingDO, timeout=400):
        data = {
            "condition": condition.dict(),
            "setting": setting.dict()
        }
        data = json.dumps(data)
        response=requests.post(f'{self.url}/instances', 
            headers=self.header, data=data, timeout=timeout)
        infos = json.loads(response.text)
        self._check_error(response.status_code, infos)
        return [InstanceInfoDO(**info) for info in infos]

    def get_instance_by_name(self, region_id: str, name: str, timeout=10):
        response=requests.get(f'{self.url}/instances/region_id/{region_id}/name/{name}', 
            headers=self.header, timeout=timeout)
        infos = json.loads(response.text)
        self._check_error(response.status_code, infos)
        return [InstanceInfoDO(**info) for info in infos]

    def release_instances(self, release_instance_infos: List[ReleaseInstanceInfoDO], timeout=10):
        instance_infos = [info.dict() for info in release_instance_infos]
        data = json.dumps(instance_infos)
        response=requests.delete(f'{self.url}/instances', 
            headers=self.header, data=data, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return info

    def run_command(
        self, 
        hosts: List[CommandHostDO],
        commands: List[CommandSettingDetailDO],
        context: Optional[CommandContextDO]=None,
    ):
        timeout = 5
        for c in commands:
            timeout += c.timeout
        hosts = [h.dict() for h in hosts]
        commands = [c.dict() for c in commands]
        if context is None:
            context = CommandContextDO()
        data = {
            "hosts": hosts,
            "commands": commands,
            "context": context.dict()
        }
        data = json.dumps(data)
        response=requests.post(f'{self.url}/instances/command', 
            headers=self.header, data=data, timeout=timeout)
        infos = json.loads(response.text)
        self._check_error(response.status_code, infos)
        return [CommandResponseDO(**r) for r in infos]


    def create_dns_record(
        self,
        record: DNSRecordDO,
        timeout=30
    ):
        data = record.to_json()
        data = json.dumps(data)
        response = requests.post(f'{self.url}/dns_record',
            headers = self.header, data=data, timeout=timeout
        )
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return DNSRecordDO(**info)

    def get_dns_records(
        self,
        domain_name: str,
        timeout=20
    ):
        response = requests.get(f'{self.url}/dns_records/domain_name/{domain_name}', 
            headers=self.header, timeout=timeout)
        infos = json.loads(response.text)
        self._check_error(response.status_code, infos)
        return [DNSRecordDO(**info) for info in infos]

    def update_dns_record(self, record: DNSRecordDO, timeout=30):
        data = json.dumps(record.to_json())
        response = requests.put(f'{self.url}/dns_record', 
            headers=self.header, data=data, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return DNSRecordDO(**info)

    def delete_dns_record(self, record_id: str, timeout=30):
        response = requests.delete(f'{self.url}/dns_record/record_id/{record_id}',
            headers=self.header, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return info

    def send_instance_creation_request(self, request: InstanceCreationRequestDO, timeout=3):
        data = json.dumps(request.dict())
        response=requests.post(f'{self.url}/instance/request', 
            headers=self.header, data=data, timeout=timeout)
        request_id = json.loads(response.text)
        self._check_error(response.status_code, request_id)
        return request_id

    def find_instance_creation_item(self, id: str, timeout=3):
        response=requests.get(f'{self.url}/instance/item/id/{id}', 
            headers=self.header, timeout=timeout)
        item = json.loads(response.text)
        self._check_error(response.status_code, item)
        return InstanceCreationItemDO(**item)

    def find_unprocessed_instance_creation_item(self, timeout=3):
        response=requests.get(f'{self.url}/instance/item/unprocessed', 
            headers=self.header, timeout=timeout)
        item = json.loads(response.text)
        self._check_error(response.status_code, item)
        if item is None:
            return None
        else:
            return InstanceCreationItemDO(**item)

    def update_instance_creation_item(self, item:InstanceCreationItemDO, timeout=10):
        data = json.dumps(item.dict())
        response=requests.put(f'{self.url}/instance/item', 
            headers=self.header, data=data, timeout=timeout)
        succeed = json.loads(response.text)
        self._check_error(response.status_code, succeed)
        return succeed

    def clear_instance_creation_item(self, timeout=3):
        response=requests.get(f'{self.url}/instance/item/clear', 
            headers=self.header, timeout=timeout)
        n = json.loads(response.text)
        self._check_error(response.status_code, n)
        return n

    def delete_instance_creation_item(self, item_id:str, timeout=3):
        response=requests.delete(f'{self.url}/instance/item/id/{item_id}', 
            headers=self.header, timeout=timeout)
        succeed = json.loads(response.text)
        self._check_error(response.status_code, succeed)
        return succeed

    def send_ban_instance_type_request(self, request: BanInstanceTypeRequestDO, timeout=3):
        data = json.dumps(request.dict())
        response=requests.post(f'{self.url}/instance_type/ban_queue/request', 
            headers=self.header, data=data, timeout=timeout)
        succeed = json.loads(response.text)
        self._check_error(response.status_code, succeed)
        return succeed

    def find_banned_instance_types(self, region_id:str, zone_id:str, timeout=3):
        response=requests.get(f'{self.url}/instance_type/ban_queue/region_id/{region_id}/zone_id/{zone_id}', 
            headers=self.header, timeout=timeout)
        item = json.loads(response.text)
        self._check_error(response.status_code, item)
        return BannedInstanceTypeResponseDO(**item)

    def check_instance_type_banned(self, request: BanInstanceTypeRequestDO, timeout=3)->Optional[bool]:
        data = json.dumps(request.dict())
        response=requests.get(f'{self.url}/instance_type/check_banned', 
            headers=self.header, data=data, timeout=timeout)
        banned = json.loads(response.text)
        self._check_error(response.status_code, banned)
        return banned

    def get_command_queue_info(self, timeout=3):
        response=requests.get(f'{self.url}/command/queue/info', 
            headers=self.header, timeout=timeout)
        info = json.loads(response.text)
        self._check_error(response.status_code, info)
        return info

    def send_command_request(self, request: CommandRequestDO, timeout=3):
        data = json.dumps(request.dict())
        response=requests.post(f'{self.url}/command/queue/request', 
            headers=self.header, data=data, timeout=timeout)
        request_id = json.loads(response.text)
        self._check_error(response.status_code, request_id)
        return request_id

    def find_command_item(self, id: str, timeout=3):
        response=requests.get(f'{self.url}/command/queue/id/{id}', 
            headers=self.header, timeout=timeout)
        item = json.loads(response.text)
        self._check_error(response.status_code, item)
        return CommandItemDO(**item)

    def find_unprocessed_command_item(self, timeout=3):
        response=requests.get(f'{self.url}/command/queue/unprocessed', 
            headers=self.header, timeout=timeout)
        item = json.loads(response.text)
        self._check_error(response.status_code, item)
        if item is None:
            return None
        else:
            return CommandItemDO(**item)

    def update_command_item(self, item:CommandItemDO, timeout=10):
        data = json.dumps(item.dict())
        response=requests.put(f'{self.url}/command/queue', 
            headers=self.header, data=data, timeout=timeout)
        succeed = json.loads(response.text)
        self._check_error(response.status_code, succeed)
        return succeed

    def clear_command_item(self, timeout=3):
        response=requests.get(f'{self.url}/command/queue/clear', 
            headers=self.header, timeout=timeout)
        n = json.loads(response.text)
        self._check_error(response.status_code, n)
        return n

    def delete_command_item(self, item_id:str, timeout=3):
        response=requests.delete(f'{self.url}/command/queue/id/{item_id}', 
            headers=self.header, timeout=timeout)
        succeed = json.loads(response.text)
        self._check_error(response.status_code, succeed)
        return succeed



from typing import Dict
from ddd_objects.infrastructure.repository_impl import RepositoryImpl
from ..domain.repository import CommandRepository
from .ao import VPCController
from ..settings import vpc_ip, vpc_port, vpc_token
from ..domain.entity import CommandRequest, CommandItem
from ..domain.value_obj import CommandID
from .converter import command_request_converter, command_item_converter

vpc_ao = VPCController(vpc_ip, vpc_port, vpc_token)
class CommandRepositoryImpl(CommandRepository, RepositoryImpl):
    def get_queue_info(self)->Dict:
        return vpc_ao.get_command_queue_info()
    
    def send_request(self, request: CommandRequest) -> CommandID:
        request = command_request_converter.to_do(request)
        request_id = vpc_ao.send_command_request(request)
        return CommandID(request_id)
    
    def find_item(self, request_id: CommandID) -> CommandItem:
        request_id = request_id.get_value()
        item = vpc_ao.find_command_item(request_id)
        return command_item_converter.to_entity(item)
command_repo = CommandRepositoryImpl()



from dataclasses import dataclass
from typing import Dict, List, Optional, Union
from pydantic import BaseModel
from ddd_objects.infrastructure.do import BaseDO, ExpiredDO


class BanInstanceTypeRequestDO(BaseModel):
    instance_type: str
    region_id: str
    zone_id : str
    duration: Optional[int] = None


class BannedInstanceTypeResponseDO(BaseModel):
    instance_types: List[str]
    region_id: str
    zone_id: Optional[str]


class ConditionDO(BaseModel):
    min_cpu_num: Optional[int] = 1
    max_cpu_num: Optional[int] = 1
    min_memory_size: Optional[int] = 1
    max_memory_size: Optional[int] = 1
    min_gpu_num: Optional[int] = None
    max_gpu_num: Optional[int] = None
    min_gpu_memory_size: Optional[int] = None
    max_gpu_memory_size: Optional[int] = None

class InstanceUserSettingDO(BaseModel):
    name: str
    password: str = 'Abcd1234'
    amount: int = 1
    image_id: Optional[str]
    region_id: str
    internet_pay_type: Optional[str]
    bandwidth_in: int = 200
    bandwidth_out: int = 1
    user_data: Optional[str] = None
    disk_size: int = 20
    key_name: str
    exclude_instance_types: List[str]=[]
    inner_connection: bool = True

class InstanceCreationRequestDO(BaseModel):
    instance_user_setting: InstanceUserSettingDO
    condition: ConditionDO
    priority: int = 3
    timeout: int = 400

class InstanceInfoDO(BaseModel):
    id: str
    instance_type: str
    create_time: str
    name: str
    hostname: str
    pay_type: str
    public_ip: List[str]
    private_ip: Optional[str]
    os_name: str
    price: float
    image_id: str
    region_id: str
    zone_id: str
    internet_pay_type: str
    bandwidth_in: int
    bandwidth_out: int
    security_group_id: List[str]
    instance_expired_time: Optional[str]
    auto_release_time: Optional[str]
    status: str
    key_name: str
    _life_time: int = 5

class InstanceCreationItemDO(BaseModel):
    id: str
    instance_creation_request: Optional[InstanceCreationRequestDO]
    status: str
    creation_time: str
    details: Optional[List[InstanceInfoDO]] = None
    entry_time: Optional[str] = None
    exit_time: Optional[str] = None
    exception: Optional[str] = None
    _life_time: int=86400

class InstanceInfosDO(ExpiredDO):
    def __init__(self, instance_infos: List[InstanceInfoDO]) -> None:
        self.instance_infos = instance_infos
        super().__init__(life_time=5)

class InstanceSettingDO(ExpiredDO):
    def __init__(
        self,
        region_id: str,
        zone_id: str,
        spot_price: float,
        instance_type: str,
        amount: int,
        image_id: Optional[str],
        instance_name: str,
        password: Optional[str],
        disk_size: int,
        bandwidth_out: int,
        bandwidth_in: int,
        user_data: Optional[str],
        internet_pay_type: Optional[str],
        security_group_id: List[str],
        key_name: str,
    ):
        self.region_id=region_id
        self.zone_id=zone_id
        self.spot_price=spot_price
        self.instance_type=instance_type
        self.amount=amount
        self.image_id=image_id
        self.instance_name=instance_name
        self.password=password
        self.disk_size=disk_size
        self.bandwidth_out=bandwidth_out
        self.bandwidth_in=bandwidth_in
        self.user_data=user_data
        self.internet_pay_type=internet_pay_type
        self.security_group_id=security_group_id
        self.key_name = key_name
        super().__init__(life_time=5)

class InstanceTypeMetaDO(BaseModel):
    instance_type_meta_id: str
    CPU_number: float
    memory_size: float
    GPU_type:str
    GPU_number:int

class InstanceTypesMetaDO(ExpiredDO):
    def __init__(
        self,
        instance_types_meta = List[InstanceTypeMetaDO]
    ):
        self.instance_types_meta = instance_types_meta
        super().__init__(life_time=86400)

class InstanceTypeStatusInfoDO(ExpiredDO):
    def __init__(
        self,
        status: str,
        status_category: str
    ):
        self.status=status
        self.status_category=status_category
        super().__init__(life_time=30)

class RegionMetaDO(ExpiredDO):
    def __init__(
        self,
        name: str,
        id: str,
        endpoint: str
    ):
        self.name=name
        self.id=id
        self.endpoint=endpoint
        super().__init__(life_time=86400)

class RegionsMetaDO(ExpiredDO):
    def __init__(
        self,
        regions_meta: List[RegionMetaDO]
    ):
        self.regions_meta=regions_meta
        super().__init__(life_time=86400)

class SecurityGroupDO(BaseModel):
        id: str
        name: str

class SpotPriceDO(ExpiredDO):
    def __init__(
        self,
        spot_price: float
    ):
        self.spot_price=spot_price
        super().__init__(life_time=1800)

class ZoneDO(BaseModel):
    name: str
    id: str
    available_instance_types: List[str]
    

class ZonesDO(ExpiredDO):
    def __init__(
        self,
        zones: List[ZoneDO]
    ):
        self.zones=zones
        super().__init__(life_time=86400)
        
class KeyInfoDO(ExpiredDO):
    def __init__(
        self,
        name: str,
        private_key: str
    ):
        self.name = name
        self.private_key = private_key
        super().__init__(life_time=86400)

class CommandContextDO(BaseModel):
    connection: str='smart'
    module_path: str='/usr/local/bin/ansible'
    forks: int= 1
    timeout: int=300 
    remote_user: str='root'
    ask_pass: bool = False 
    ssh_extra_args: Optional[str] = None
    sftp_extra_args: Optional[str] = None
    scp_extra_args: Optional[str] = None 
    ask_value_pass: bool = False
    listhosts: bool = False 
    listtasks: bool = False 
    listtags: bool = False
    syntax: bool = False
    ssh_common_args='-o StrictHostKeyChecking=no'
    become: Optional[str] = None
    become_method: Optional[str] = None
    become_user: str = 'root' 
    check: bool = False 
    diff: bool = False
    verbosity: int = 1


class CommandSettingDetailDO(BaseModel):
    command: Union[str, dict]
    module: str = 'shell'
    retries: int = 2
    timeout: int = 3
    delay: float = 0.1


class CommandHostDO(BaseModel):
    ip: str
    hostname: Optional[str] = None
    port: int = 22
    username: str = 'root'

class CommandRequestDO(BaseModel):
    commands: Union[List[str], List[CommandSettingDetailDO]]
    hosts: Union[List[str], List[CommandHostDO]]
    context: Optional[CommandContextDO] = None
    priority: int = 3
    timeout: int = 3

class CommandResponseDO(BaseModel):
    status: str
    ip: str
    message: Optional[str] = None
    exception: Optional[str] = None
    stdout: Optional[str] = None
    cmd: Optional[str] = None
    start_time: Optional[str] = None
    end_time: Optional[str] = None
    delta_time: Optional[str] = None
    entry_time: Optional[str] = None
    exit_time: Optional[str] = None
    stderr: Optional[str] = None

class CommandItemDO(BaseModel):
    id: str
    command_request: Optional[CommandRequestDO]
    creation_time: str
    status: str
    details: Optional[List[CommandResponseDO]] = None
    exception: Optional[str] = None
    _life_time: int = 600
    
class CommandUserSettingDO(BaseDO):
    def __init__(
        self,
        hostname: str,
        ip: str,
        username: str = 'root',
        port: int = 22,
        password: Optional[str] = None,
        private_key_file: Optional[str] = None,
    ) -> None:
        self.hostname = hostname
        self.ip = ip
        self.username = username
        self.password = password
        self.port = port
        self.private_key_file=private_key_file
        
class CommandUserSettingGroupDO:
    def __init__(
        self,
        group: List[CommandUserSettingDO],
    ):
        self.group = group

class CommandDO:
    def __init__(self, args:str, module: str='shell'):
        self.args = args
        self.module = module

class CommandResultDO(ExpiredDO):
    def __init__(
        self, 
        succeed: bool,
        output: str,
        ip: str
    ):
        self.succeed = succeed
        self.output = output
        self.ip = ip
        super().__init__(life_time=1)

class CommandResultsDO(ExpiredDO):
    def __init__(
        self,
        results: Dict[str, CommandResultDO]
    ):
        self.results = results

@dataclass
class DNSRecordDO:
    domain_name: str
    subdomain: str
    value: str
    id: Optional[str] = None
    dns_type: str = 'A'
    weight: Optional[int] = None
    ttl: int = 600
    priority: Optional[int] = None
    line: Optional[str] = None

@dataclass
class SecurityGroupPermissionDO(BaseDO):
    region_id: str
    security_group_id: str
    port: int
    dest_cidr_ip: Optional[str]=None
    source_cidr_ip: Optional[str]=None
    direction: str='egress'
    nic_type: str='internet'
    policy: str='Accept'
    description: str=''
    ip_protocol: str='TCP'

@dataclass
class DNSRecordDO(BaseDO):
    domain_name: str
    subdomain: str
    value: str
    id: Optional[str]=None
    weight: Optional[int]=None
    dns_type: str='A'
    ttl: int=600
    priority: Optional[int]=None
    line: Optional[str]=None

class ReleaseInstanceInfoDO(BaseModel):
    region_id: str
    instance_id: str

